import pandas as pd


def _check_necessary_opf_parameters(net, logger):
    # Check if all necessary parameters are given:
    opf_col = {
        'ext_grid': pd.Series(['min_p_mw', 'max_p_mw', 'min_q_mvar', 'max_q_mvar']),
        'gen': pd.Series(['min_p_mw', 'max_p_mw', 'min_q_mvar', 'max_q_mvar']),
        'sgen': pd.Series(['min_p_mw', 'max_p_mw', 'min_q_mvar', 'max_q_mvar']),
        'load': pd.Series(['min_p_mw', 'max_p_mw', 'min_q_mvar', 'max_q_mvar']),
        'storage': pd.Series(['min_p_mw', 'max_p_mw', 'min_q_mvar', 'max_q_mvar']),
        'dcline': pd.Series(['max_p_mw', 'min_q_from_mvar', 'min_q_to_mvar', 'max_q_from_mvar',
                             'max_q_to_mvar'])}
    missing_val = []
    error = False
    for element_type, columns in opf_col.items():
        if len(net[element_type]):

            # --- determine controllables
            if element_type in ["ext_grid", "dcline"]:
                controllables = net[element_type].index
            else:
                if "controllable" in net[element_type].columns:
                    net[element_type].controllable = net[element_type].controllable.fillna(element_type == 'gen')
                    controllables = net[element_type].index[net[element_type].controllable.astype(
                        bool)]
                else:
                    controllables = net[element_type].index if element_type == 'gen' else []

            # --- logging for missing data in element tables with controllables
            if len(controllables):

                missing_col = columns.loc[~columns.isin(net[element_type].columns)].values
                na_col = [col for col in set(columns)-set(missing_col) if
                          net[element_type][col].loc[controllables].isnull().any()]

                if len(missing_col):
                    if element_type != "ext_grid":
                        logger.error("These columns are missing in " + element_type + ": " +
                                     str(missing_col))
                        error = True
                    else:  # "ext_grid" -> no error due to missing columns at ext_grid
                        logger.debug("These missing columns in ext_grid are considered in OPF as " +
                                     "+- 1000 TW.: " + str(missing_col))

                # determine missing values
                if len(na_col):
                    missing_val.append(element_type)

    if missing_val:
        logger.info("These elements have missing power constraint values, which are considered " +
                    "in OPF as +- 1000 TW: " + str(missing_val))

    # voltage limits: no error due to missing voltage limits
    if 'min_vm_pu' in net.bus.columns:
        if net.bus.min_vm_pu.isnull().any():
            logger.info("There are missing bus.min_vm_pu values, which are considered in OPF as " +
                        "0.0 pu.")
    else:
        logger.info("'min_vm_pu' is missing in bus table. In OPF these limits are considered as " +
                    "0.0 pu.")
    if 'max_vm_pu' in net.bus.columns:
        if net.bus.max_vm_pu.isnull().any():
            logger.info("There are missing bus.max_vm_pu values, which are considered in OPF as " +
                        "2.0 pu.")
    else:
        logger.info("'max_vm_pu' is missing in bus table. In OPF these limits are considered as " +
                    "2.0 pu.")

    if error:
        raise KeyError("OPF parameters are not set correctly. See error log.")

    # --- log multiple costs to elements
    cost_check_df = pd.concat([net.poly_cost[["element", "et"]], net.poly_cost[["element", "et"]]],
                              ignore_index=True)
    cost_check_df["power_type"] = ["p"]*net.poly_cost.shape[0] + ["q"]*net.poly_cost.shape[0]
    cost_check_df = pd.concat([cost_check_df, net.pwl_cost[["element", "et", "power_type"]]],
                              ignore_index=True)
    if cost_check_df.duplicated().any():
        raise UserWarning("There are multiple costs to one or multiple elements.")



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
import sys
try:
    import matplotlib.pyplot as plt
    MATPLOTLIB_INSTALLED = True
except ImportError:
    MATPLOTLIB_INSTALLED = False

from pandapower.auxiliary import soft_dependency_error
from pandapower.plotting.plotting_toolbox import get_collection_sizes
from pandapower.plotting.collections import create_bus_collection, create_line_collection, \
    create_trafo_collection, create_trafo3w_collection, \
    create_line_switch_collection, draw_collections, create_bus_bus_switch_collection, \
    create_ext_grid_collection, create_sgen_collection, \
    create_gen_collection, create_load_collection, create_dcline_collection, create_vsc_collection
from pandapower.plotting.generic_geodata import create_generic_coordinates

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def simple_plot(net, respect_switches=False, line_width=1.0, bus_size=1.0, ext_grid_size=1.0,
                trafo_size=1.0, plot_loads=False, plot_gens=False, plot_sgens=False, load_size=1.0,
                gen_size=1.0, sgen_size=1.0,
                switch_size=2.0, switch_distance=1.0, plot_line_switches=False, scale_size=True,
                bus_color='b', line_color='grey',  dcline_color='c', trafo_color='k',
                ext_grid_color='y', switch_color='k', library='igraph', show_plot=True, ax=None,
                bus_dc_size=1.0, bus_dc_color="m", line_dc_color="c", vsc_size=4.0, vsc_color="orange"):
    """
        Plots a pandapower network as simple as possible. If no geodata is available, artificial
        geodata is generated. For advanced plotting see the tutorial

        INPUT:
            **net** - The pandapower format network.

        OPTIONAL:
            **respect_switches** (bool, False) - Respect switches if artificial geodata is created.
                                                This Flag is ignored if plot_line_switches is True

            **line_width** (float, 1.0) - width of lines

            **bus_size** (float, 1.0) - Relative size of buses to plot.
                                        The value bus_size is multiplied with mean_distance_between_buses, which equals
                                        the distance between the max geocoord and the min divided by 200.
                                        mean_distance_between_buses = sum((net.bus.geo.max() - net.bus.geo.min()) / 200)

            **ext_grid_size** (float, 1.0) - Relative size of ext_grids to plot. See bus sizes for details.
                                                Note: ext_grids are plotted as rectangles

            **trafo_size** (float, 1.0) - Relative size of trafos to plot.

            **plot_loads** (bool, False) - Flag to decide whether load symbols should be drawn.

            **plot_gens** (bool, False) - Flag to decide whether gen symbols should be drawn.

            **plot_sgens** (bool, False) - Flag to decide whether sgen symbols should be drawn.

            **load_size** (float, 1.0) - Relative size of loads to plot.

            **sgen_size** (float, 1.0) - Relative size of sgens to plot.

            **switch_size** (float, 2.0) - Relative size of switches to plot. See bus size for details

            **switch_distance** (float, 1.0) - Relative distance of the switch to its corresponding \
                                               bus. See bus size for details

            **plot_line_switches** (bool, False) - Flag if line switches are plotted

            **scale_size** (bool, True) - Flag if bus_size, ext_grid_size, bus_size- and distance \
                                          will be scaled with respect to grid mean distances

            **bus_color** (String, colors[0]) - Bus Color. Init as first value of color palette. Usually colors[0] = "b".

            **line_color** (String, 'grey') - Line Color. Init is grey

            **dcline_color** (String, 'c') - Line Color. Init is cyan

            **trafo_color** (String, 'k') - Trafo Color. Init is black

            **ext_grid_color** (String, 'y') - External Grid Color. Init is yellow

            **switch_color** (String, 'k') - Switch Color. Init is black

            **library** (String, "igraph") - library name to create generic coordinates (case of
                                                missing geodata). "igraph" to use igraph package or "networkx" to use networkx package.

            **show_plot** (bool, True) - Shows plot at the end of plotting

            **ax** (object, None) - matplotlib axis to plot to

        OUTPUT:
            **ax** - axes of figure
    """
    try:
        if hasattr(net, "bus_geodata") or hasattr(net, "line_geodata"):
            raise UserWarning("""The supplied network uses an outdated geodata format. Please update your geodata by
                                 \rrunning `pandapower.plotting.geo.convert_geodata_to_geojson(net)`""")
    except UserWarning as e:
        logger.warning(e)

    # don't hide lines if switches are plotted
    if plot_line_switches:
        respect_switches = False

    # create geocoord if none are available
    if (len(net.line.geo) == 0 and len(net.bus.geo) == 0) or (net.line.geo.isna().any() and net.bus.geo.isna().any()):
        logger.warning("No or insufficient geodata available --> Creating artificial coordinates." +
                       " This may take some time")
        create_generic_coordinates(net, respect_switches=respect_switches, library=library)

    if scale_size:
        # if scale_size -> calc size from distance between min and max geocoord
        sizes = get_collection_sizes(net, bus_size, ext_grid_size, trafo_size,
                                     load_size, sgen_size, switch_size, switch_distance, gen_size)
        bus_size = sizes["bus"]
        ext_grid_size = sizes["ext_grid"]
        trafo_size = sizes["trafo"]
        sgen_size = sizes["sgen"]
        load_size = sizes["load"]
        switch_size = sizes["switch"]
        switch_distance = sizes["switch_distance"]
        gen_size = sizes["gen"]

    # create bus collections to plot
    bc = create_bus_collection(net, net.bus.index, size=bus_size, color=bus_color, zorder=10)

    # if bus geodata is available, but no line geodata
    use_bus_geodata = len(net.line.geo.dropna()) == 0
    in_service_lines = net.line[net.line.in_service].index
    nogolines = set(net.switch.element[(net.switch.et == "l") & (net.switch.closed == 0)]) \
        if respect_switches else set()
    plot_lines = in_service_lines.difference(nogolines)
    plot_dclines = net.dcline.in_service
    plot_lines_dc = net.line_dc.loc[net.line_dc.in_service].index

    # create line collections
    lc = create_line_collection(net, plot_lines, color=line_color, linewidths=line_width,
                                use_bus_geodata=use_bus_geodata)
    collections = [bc, lc]

    # create dcline collections
    if len(net.dcline) > 0:
        dclc = create_dcline_collection(net, plot_dclines, color=dcline_color,
                                        linewidths=line_width)
        collections.append(dclc)
    # create bus dc collection
    if len(net.bus_dc) > 0:
        bc_dc = create_bus_collection(net, net.bus_dc.index, size=bus_dc_size, color=bus_dc_color, zorder=10,
                                      bus_table="bus_dc")
        collections.append(bc_dc)
    # create VSC collection
    if len(net.vsc) > 0:
        vsc_ac = create_vsc_collection(net, net.vsc.index, size=vsc_size, color=vsc_color, zorder=12)
        collections.append(vsc_ac)
    # create line_dc collections
    if len(net.line_dc) > 0:
        lc_dc = create_line_collection(net, plot_lines_dc, color=line_dc_color, linewidths=line_width,
                                       use_bus_geodata=use_bus_geodata, line_table="line_dc")
        collections.append(lc_dc)

    # create ext_grid collections
    # eg_buses_with_geo_coordinates = set(net.ext_grid.bus.values) & set(net.bus_geodata.index)
    if len(net.ext_grid) > 0:
        sc = create_ext_grid_collection(net, size=ext_grid_size, orientation=0,
                                        ext_grids=net.ext_grid.index, patch_edgecolor=ext_grid_color,
                                        zorder=11)
        collections.append(sc)

    # create trafo collection if trafo is available
    trafo_buses_with_geo_coordinates = [t for t, trafo in net.trafo.iterrows()
                                        if trafo.hv_bus in net.bus.geo.index and
                                        trafo.lv_bus in net.bus.geo.index]
    if len(trafo_buses_with_geo_coordinates) > 0:
        tc = create_trafo_collection(net, trafo_buses_with_geo_coordinates,
                                     color=trafo_color, size=trafo_size)
        collections.append(tc)

    # create trafo3w collection if trafo3w is available
    trafo3w_buses_with_geo_coordinates = [
        t for t, trafo3w in net.trafo3w.iterrows() if trafo3w.hv_bus in net.bus.geo.index and
                                                      trafo3w.mv_bus in net.bus.geo.index and
                                                      trafo3w.lv_bus in net.bus.geo.index]
    if len(trafo3w_buses_with_geo_coordinates) > 0:
        tc = create_trafo3w_collection(net, trafo3w_buses_with_geo_coordinates,
                                       color=trafo_color)
        collections.append(tc)

    if plot_line_switches and len(net.switch):
        sc = create_line_switch_collection(
            net, size=switch_size, distance_to_bus=switch_distance,
            use_line_geodata=not use_bus_geodata, zorder=12, color=switch_color)
        collections.append(sc)

    if plot_sgens and len(net.sgen):
        sgc = create_sgen_collection(net, size=sgen_size, orientation=0)
        collections.append(sgc)
    if plot_gens and len(net.gen):
        gc = create_gen_collection(net, size=gen_size)
        collections.append(gc)
    if plot_loads and len(net.load):
        lc = create_load_collection(net, size=load_size)
        collections.append(lc)

    if len(net.switch):
        bsc = create_bus_bus_switch_collection(net, size=switch_size)
        collections.append(bsc)

    ax = draw_collections(collections, ax=ax)
    if show_plot:
        if not MATPLOTLIB_INSTALLED:
            soft_dependency_error(str(sys._getframe().f_code.co_name) + "()", "matplotlib")
        plt.show()
    return ax


if __name__ == "__main__":
    from pandapower.networks.power_system_test_cases import case145
    # from pandapower.networks.cigre_networks import create_cigre_network_mv
    # from pandapower.networks.mv_oberrhein import mv_oberrhein

    net = case145()
    #    net = nw.create_cigre_network_mv()
    #    net = nw.mv_oberrhein()
    simple_plot(net, bus_size=0.4)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import sys
from itertools import combinations
import warnings

import numpy as np
import pandas as pd
import networkx as nx
try:
    import matplotlib.pyplot as plt
    MATPLOTLIB_INSTALLED = True
except ImportError:
    MATPLOTLIB_INSTALLED = False

from pandapower.auxiliary import soft_dependency_error, warn_and_fix_parameter_renaming
from pandapower.topology.create_graph import create_nxgraph
from pandapower.topology.graph_searches import calc_distance_to_bus


def plot_voltage_profile(net, ax=None, plot_transformers=True, xlabel="Distance from Slack [km]",
                         ylabel="Voltage [pu]", x0=0, line_color="grey", trafo_color="r",
                         bus_colors="b", line_loading_weight=False, voltage_column=None,
                         bus_size=3, lines=None, **kwargs):
    """Plot the voltage profile depending on the distance from the slack.

    Parameters
    ----------
    net : pp.PandapowerNet
        net including power flow results
    ax : matplotlib.axes, optional
        axis to plot to, by default None
    plot_transformers : bool, optional
        Whether vertical lines should be plotted to display the voltage drop of the transformers,
        by default True
    xlabel : str, optional
        xlable of the figure, by default "Distance from Slack [km]"
    ylabel : str, optional
        ylable of the figure, by default "Voltage [pu]"
    x0 : int, optional
        slack position at the xaxis, by default 0
    line_color : str, optional
        color used to plot the lines, by default "grey"
    trafo_color : str, optional
        color used to plot the trafos, by default "r"
    bus_colors : [str, dict[int, str]], optional
        colors used to plot the buses. Can be passed as string (to give all buses the same color),
        or as dict, by default "b"
    line_loading_weight : bool, optional
        enables loading dependent width of the lines, by default False
    bus_size : int, optional
        size of bus representations, by default 3
    lines : Any[list[int], pd.Index[int]], optional
        list of line indices which should be plottet. If None, all lines are plotted, by default None

    Returns
    -------
    matplotlib.axes
        axis of the plot

    """
    # handle exceptions and inputs
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name)+"()", "matplotlib")
    if "voltage_column" in kwargs:
        raise DeprecationWarning("Parameter 'voltage_column' has been removed.")
    trafo_color = warn_and_fix_parameter_renaming(
        "trafocolor", "trafo_color", trafo_color, "r", **kwargs)
    if ax is None:
        plt.figure(facecolor="white", dpi=120)
        ax = plt.gca()
    if not net.converged and not net.OPF_converged:
        raise ValueError("no results in this pandapower network")
    if lines is None:
        lines = net.line.index

    # run plotting code
    sl_buses = np.union1d(
        net.ext_grid.loc[net.ext_grid.in_service, "bus"].values,
        net.gen.loc[net.gen.slack & net.gen.in_service, "bus"].values)
    for eg in sl_buses:
        d = calc_distance_to_bus(net, eg)
        for lix, line in net.line[net.line.in_service & net.line.index.isin(lines)].iterrows():
            if line.from_bus not in d.index:
                continue
            if not ((net.switch.element == line.name) & ~net.switch.closed & (
                    net.switch.et == 'l')).any():
                from_bus = line.from_bus
                to_bus = line.to_bus
                x = [x0 + d.at[from_bus], x0 + d.at[to_bus]]
                try:
                    y = [net.res_bus.vm_pu.at[from_bus], net.res_bus.vm_pu.at[to_bus]]
                except:
                    raise UserWarning
                if "linewidth" in kwargs or not line_loading_weight:
                    ax.plot(x, y, color=line_color, **kwargs)
                else:
                    ax.plot(x, y, linewidth=0.4 * np.sqrt(net.res_line.loading_percent.at[lix]),
                            color=line_color, **kwargs)
                if bus_colors is not None:
                    if isinstance(bus_colors, str):
                        bus_colors = {b: bus_colors for b in net.bus.index}
                    for bus, x, y in zip((from_bus, to_bus), x, y):
                        if bus in bus_colors:
                            ax.plot(x, y, 'o', color=bus_colors[bus], ms=bus_size)
                kwargs = {k: v for k, v in kwargs.items() if not k == "label"}

        # trafo geodata
        if plot_transformers:
            for trafo_table in ['trafo', 'trafo3w']:
                if trafo_table not in net.keys():
                    continue
                transformers = net[trafo_table].query('in_service')
                for tid, tr in transformers.iterrows():
                    t_buses = [tr[b_col] for b_col in ('lv_bus', 'mv_bus', 'hv_bus') if
                               b_col in tr.index]
                    if any([b not in d.index.values or b not in net.res_bus.index.values for b in
                            t_buses]):
                        # logger.info('cannot add trafo %d to plot' % tid)
                        continue

                    for bi, bj in combinations(t_buses, 2):
                        tr_coords = ([x0 + d.loc[bi], x0 + d.loc[bj]],
                                     [net.res_bus.at[bi, 'vm_pu'], net.res_bus.at[bj, 'vm_pu']])
                        ax.plot(*tr_coords, color=trafo_color,
                                **{k: v for k, v in kwargs.items() if not k == "color"})

        if xlabel:
            ax.set_xlabel(xlabel, fontweight="bold", color=(.4, .4, .4))
        if ylabel:
            ax.set_ylabel(ylabel, fontweight="bold", color=(.4, .4, .4))
    return ax


def plot_loading(net, ax=None, element_type="line", box_color="b", median_color="r",
                 whisker_color="k", index_subset=None, **kwargs):
    """Plot a boxplot of loadings

    Parameters
    ----------
    net : pp.PandapowerNet
        net including power flow results
    ax : matplotlib.axes, optional
        axis to plot to, by default None
    element : str, optional
        name of element which loadings should be plotted, by default "line"
    box_color : str, optional
        color of the box, by default "b"
    median_color : str, optional
        color of the median line, by default "r"
    whisker_color : str, optional
        color of the whiskers, by default "k"
    index_subset : Any[list[int], pd.Index[int]], optional
        list of element indices which should be considered. If None, all elements are considered,
        by default None

    Returns
    -------
    matplotlib.axes
        axis of the plot
    """
    # handle exceptions and inputs
    deprecated = [("boxcolor", "box_color", "b"),
                  ("mediancolor", "median_color", "r"),
                  ("whiskercolor", "whisker_color", "k"),
                  ("element", "element_type", "line")]
    for (old, new, default) in deprecated:
        locals()[new] = warn_and_fix_parameter_renaming(
            old, new, locals()[new], default, **kwargs)
    if not MATPLOTLIB_INSTALLED:
        soft_dependency_error(str(sys._getframe().f_code.co_name)+"()", "matplotlib")
    if ax is None:
        plt.figure(facecolor="white", dpi=80)
        ax = plt.gca()
    if index_subset is None:
        index_subset = net[element_type].index

    # run plotting code
    loadings = net["res_%s" % element_type].loading_percent.values[net[
        "res_%s" % element_type].index.isin(index_subset)]
    boxplot = ax.boxplot(loadings[~np.isnan(loadings)], whis=[0, 100])
    for l in list(boxplot.keys()):
        plt.setp(boxplot[l], lw=3)
        if l == "medians":
            plt.setp(boxplot[l], color=median_color)
        elif l == "boxes" or l == "whiskers":
            plt.setp(boxplot[l], color=box_color)
        else:
            plt.setp(boxplot[l], color=whisker_color)
    ax.set_ylabel(f"{element_type.capitalize()} Loading [%]", fontweight="bold", color=(.4, .4, .4))
    ax.set_xticks([1], [""])
    return ax


def voltage_profile_to_bus_geodata(net, voltages=None, root_bus=None):
    if voltages is None:
        if not net.converged:
            raise ValueError("no results in this pandapower network")
        voltages = net.res_bus.vm_pu

    mg = create_nxgraph(net, respect_switches=True)
    sl_buses = np.r_[
        net.ext_grid.loc[net.ext_grid.in_service, "bus"].values,
        net.gen.loc[net.gen.slack & net.gen.in_service, "bus"].values]
    first_eg = sl_buses[0] if root_bus is None else root_bus
    other_eg = np.setdiff1d(sl_buses, first_eg)
    mg.add_edges_from([(first_eg, y, {"weight": 0}) for y in other_eg])
    dist = pd.Series(nx.single_source_dijkstra_path_length(mg, first_eg))

    bgd = pd.DataFrame({"x": dist.loc[net.bus.index.values].values,
                        "y": voltages.loc[net.bus.index.values].values},
                       index=net.bus.index)
    return bgd


if __name__ == "__main__":
    from pandapower.run import runpp
    from pandapower.networks.mv_oberrhein import mv_oberrhein

    net = mv_oberrhein()
    runpp(net)

    fig, axs = plt.subplots(ncols=2, figsize=(8, 5), gridspec_kw={"width_ratios": [4, 1]})
    plot_voltage_profile(net, ax=axs[0])
    plot_loading(net, ax=axs[1])
    plt.show()



# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


"""Updates bus, gen, branch data structures to match power flow soln.
"""

from numpy import asarray, angle, pi, conj, zeros, ones, finfo, c_, ix_, real, flatnonzero as find, \
    setdiff1d, intersect1d, r_, isin, arange, flatnonzero, nan_to_num, int64
from scipy.sparse import csr_matrix

from pandapower.pypower.idx_brch import F_BUS, T_BUS, BR_STATUS, PF, PT, QF, QT
from pandapower.pypower.idx_bus import VM, VA, PD, QD
from pandapower.pypower.idx_gen import GEN_BUS, GEN_STATUS, PG, QG, QMIN, QMAX, SL_FAC

EPS = finfo(float).eps


def pfsoln(baseMVA, bus0, gen0, branch0, svc, tcsc, ssc, vsc, Ybus, Yf, Yt, V, ref, ref_gens, Ibus=None, limited_gens=None):
    """Updates bus, gen, branch data structures to match power flow soln.

    @author: Ray Zimmerman (PSERC Cornell)
    @author: Richard Lincoln
    """
    # initialize return values
    bus = bus0
    gen = gen0
    branch = branch0

    # ----- update Qg for all gens and Pg for slack bus(es) -----
    # generator info
    on = find(gen[:, GEN_STATUS] > 0)  # which generators are on?
    gbus = gen[on, GEN_BUS].astype(int64)  # what buses are they at?

    # compute total injected bus powers
    Ibus = zeros(len(V)) if Ibus is None else Ibus
    Sbus = V * conj(Ybus * V - Ibus)

    _update_v(bus, V)
    _update_q(baseMVA, bus, gen, gbus, Sbus[gbus], on)

    if limited_gens is not None and len(limited_gens) > 0:
        on = find((gen[:, GEN_STATUS] > 0) | isin(arange(len(gen)), limited_gens))
        gbus = gen[on, GEN_BUS].astype(int64)

    _update_p(baseMVA, bus, gen, ref, gbus, Sbus, ref_gens)

    # ----- update/compute branch power flows -----
    out = find(branch[:, BR_STATUS] == 0)  # out-of-service branches
    br = find(branch[:, BR_STATUS]).astype(int64)  # in-service branches

    if len(out):
        raise RuntimeError
    # complex power at "from" bus
    Sf = V[real(branch[br, F_BUS]).astype(int64)] * conj(Yf[br, :] * V) * baseMVA
    # complex power injected at "to" bus
    St = V[real(branch[br, T_BUS]).astype(int64)] * conj(Yt[br, :] * V) * baseMVA
    branch[ix_(br, [PF, QF, PT, QT])] = c_[Sf.real, Sf.imag, St.real, St.imag]
    branch[ix_(out, [PF, QF, PT, QT])] = zeros((len(out), 4))

    return bus, gen, branch


def _update_v(bus, V):
    # ----- update bus voltages -----
    bus[:, VM] = abs(V)
    bus[:, VA] = angle(V) * 180. / pi


def _split_p_for_gens_at_same_bus(gen, p_bus, gens_at_bus, ref_gens):
    if len(gens_at_bus) > 1:  # more than one generator at this ref bus
        # subtract off what is generated by other gens at this bus
        ext_grids = intersect1d(gens_at_bus, ref_gens)
        pv_gens = setdiff1d(gens_at_bus, ext_grids)
        p_ext_grids = p_bus - sum(gen[pv_gens, PG])
        slack_weights = gen[ext_grids, SL_FAC]
        sum_slack_weights = sum(slack_weights)
        if sum_slack_weights > 0:
            # distribute bus slack power according to the distributed slack contribution factors
            gen[ext_grids, PG] = gen[ext_grids, PG] + (
                        p_ext_grids - sum(gen[ext_grids, PG])) * slack_weights / sum_slack_weights
        else:
            gen[ext_grids, PG] = p_ext_grids / len(ext_grids)
    else:
        gen[gens_at_bus[0], PG] = p_bus


def _update_p(baseMVA, bus, gen, ref, gbus, Sbus, ref_gens):
    # update Pg for slack bus(es)
    # inj P + local Pd
    for slack_bus in ref:
        gens_at_bus = find(gbus == slack_bus)  # which is(are) the reference gen(s)?
        # xward results come at the PQ buses and not at the PV buses:
        if len(gens_at_bus) == 0:
            bus[slack_bus, PD] = -Sbus[slack_bus].real * baseMVA  # negative because it is a PQ bus
            continue
        # p_bus = Sbus[gens_at_bus[0]].real * baseMVA + bus[slack_bus, PD]
        p_bus = Sbus[slack_bus].real * baseMVA + bus[slack_bus, PD]
        _split_p_for_gens_at_same_bus(gen, p_bus, gens_at_bus, ref_gens)


def _update_q(baseMVA, bus, gen, gbus, Sbus, on):
    # update Qg for all generators
    gen[:, QG] = zeros(gen.shape[0])  # zero out all Qg
    gen[on, QG] = Sbus.imag * baseMVA + bus[gbus, QD]  # inj Q + local Qd
    # ... at this point any buses with more than one generator will have
    # the total Q dispatch for the bus assigned to each generator. This
    # must be split between them. We do it first equally, then in proportion
    # to the reactive range of the generator.

    if len(on) > 1:
        # build connection matrix, element i, j is 1 if gen on(i) at bus j is ON
        nb = bus.shape[0]
        ngon = on.shape[0]
        Cg = csr_matrix((ones(ngon), (range(ngon), gbus)), (ngon, nb))

        # divide Qg by number of generators at the bus to distribute equally
        ngg = Cg * Cg.sum(0).T  # ngon x 1, number of gens at this gen's bus
        ngg = asarray(ngg).flatten()  # 1D array
        gen[on, QG] = gen[on, QG] / ngg

        # divide proportionally
        Cmin = csr_matrix((gen[on, QMIN], (range(ngon), gbus)), (ngon, nb))
        Cmax = csr_matrix((gen[on, QMAX], (range(ngon), gbus)), (ngon, nb))
        Qg_tot = Cg.T * gen[on, QG]  # nb x 1 vector of total Qg at each bus
        Qg_min = Cmin.sum(0).T  # nb x 1 vector of min total Qg at each bus
        Qg_max = Cmax.sum(0).T  # nb x 1 vector of max total Qg at each bus
        Qg_min = asarray(Qg_min).flatten()  # 1D array
        Qg_max = asarray(Qg_max).flatten()  # 1D array
        # gens at buses with Qg range = 0
        ig = find(Cg * Qg_min == Cg * Qg_max)
        Qg_save = gen[on[ig], QG]
        gen[on, QG] = gen[on, QMIN] + (Cg * ((Qg_tot - Qg_min) / (Qg_max - Qg_min + EPS))) * \
                      (gen[on, QMAX] - gen[on, QMIN])  # ^ avoid div by 0
        gen[on[ig], QG] = Qg_save  # (terms are mult by 0 anyway)



from pandas.api.types import is_bool_dtype, is_numeric_dtype

from pandapower.auxiliary import LoadflowNotConverged
from pandapower.create import create_buses, create_bus, create_load, create_sgen, create_gen, create_impedance
from pandapower.toolbox.grid_modification import drop_buses
from pandapower.grid_equivalents.auxiliary import (
    calc_zpbn_parameters,
    drop_internal_branch_elements,
    build_ppc_and_Ybus, drop_measurements_and_controllers,
    drop_and_edit_cost_functions, _runpp_except_voltage_angles,
    replace_motor_by_load,
    impedance_columns
)
from pandapower.grid_equivalents.toolbox import get_connected_switch_buses_groups
from copy import deepcopy
import pandas as pd
import numpy as np
import operator
import time
import re
from functools import reduce

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def _calculate_equivalent_Ybus(net_zpbn, bus_lookups, eq_type,
                               show_computing_time=False, **kwargs):
    """
    The function orders the admittance matrix of the original network into
    new format firstly, which is convenient for rei equivalent calculation.d
    Then it calculates the equivalent admittance matrix of the given network

    i: internal   b: boundary    e: external    g: ground    t: total

    Ymat_trans =

    [ Ybus_ii, Ybus_ib,     0   ]    [Ybus_ii, Ybus_ib,      0   ,    0   ,    0      ]
    [ Ybus_bi, Ybus_bb, Ybus_be ] =             _________________   _________________
    [   0   ,  Ybus_eb, Ybus_ee ]    [Ybus_bi, |Ybus_bb ,   0   |, |   0   , Ybus_be| ]
                                               |                |  |                |
                                     [   0   , |   0    ,Ybus_tt|, |Ybus_tg,    0   | ]
                                               |________________|  |________________|
                                                _________________   _________________
                                     [   0   , |   0    ,Ybus_gt|, |Ybus_gg, Ybus_ge| ]
                                     [   0   , |Ybus_eb ,    0  |, |Ybus_eg, Ybus_ee| ]
                                               |________________|  |________________|

    INPUT:
        **net_zpbn** -  zero power balance network (pandapower network)

        **bus_lookups** (dict) -  bus lookups

        **eq_type** (str) -  the equavalten type

    OPTIONAL:
        **check_validity** (bool, False) - TODO

    OUTPUT:
        **Ybus** - equivalent admittance matrix of the external network

    """
    t_start = time.perf_counter()
    # --- initialization
    Ybus_origin = net_zpbn._ppc["internal"]["Ybus"].todense()
    bus_lookup_ppc = bus_lookups["bus_lookup_ppc"]
    nb_dict = {}
    for key in bus_lookup_ppc.keys():
        if key != "b_area_buses_no_switch":
            nb_dict["nb_" + key.split("_")[0]] = len(bus_lookup_ppc[key])
    Ybus_buses = list(bus_lookup_ppc.values())
    Ybus_new_sequence = reduce(operator.concat, Ybus_buses)

    # --- transform Ybus_origin to Ybus_new according to the Ybus_new_sequence
    Ybus_sorted = Ybus_origin[:, Ybus_new_sequence][Ybus_new_sequence]
    if eq_type == "xward":
        idx_large_y = np.linspace(0, len(Ybus_new_sequence) - 1, len(Ybus_new_sequence))
        idx_large_y = ((idx_large_y >= nb_dict["nb_i"] + nb_dict["nb_b"]) &
                       (np.isin(Ybus_new_sequence[:], net_zpbn._ppc["gen"][:, 0])))
        Ybus_sorted[idx_large_y, idx_large_y] = 1e8

    # --- calculate calculate equivalent Ybus and equivalent Ybus without_internals
    Ybus_bb = Ybus_sorted[nb_dict["nb_i"]:(nb_dict["nb_i"] + nb_dict["nb_b"] + nb_dict["nb_t"]),
              nb_dict["nb_i"]:(nb_dict["nb_i"] + nb_dict["nb_b"] + nb_dict["nb_t"])]
    Ybus_ee = Ybus_sorted[-(nb_dict["nb_e"] + nb_dict["nb_g"]):,
              -(nb_dict["nb_e"] + nb_dict["nb_g"]):]
    Ybus_eb = Ybus_sorted[-(nb_dict["nb_e"] + nb_dict["nb_g"]):,
              nb_dict["nb_i"]:(nb_dict["nb_i"] + nb_dict["nb_b"] + nb_dict["nb_t"])]
    Ybus_be = Ybus_eb.T

    try:
        inverse_Ybus_ee = np.linalg.inv(Ybus_ee)
    except np.linalg.LinAlgError as err:
        if 'Singular matrix' in str(err):
            logger.debug("Ymat_ee is a singular martix, now try to compute the \
                         pseudo-inverse of the matrix.")
            inverse_Ybus_ee = np.linalg.pinv(Ybus_ee)
    Ybus_eq_boundary = Ybus_bb - (Ybus_be * inverse_Ybus_ee * Ybus_eb)
    Ybus_eq = np.copy(Ybus_sorted[0: nb_dict["nb_i"] + nb_dict["nb_b"] + nb_dict["nb_t"],
                      0: nb_dict["nb_i"] + nb_dict["nb_b"] + nb_dict["nb_t"]])
    Ybus_eq[-(nb_dict["nb_b"] + nb_dict["nb_t"]):, -(nb_dict["nb_b"] +
                                                     nb_dict["nb_t"]):] = Ybus_eq_boundary

    t_end = time.perf_counter()
    if show_computing_time:
        logger.info("\"calculate_equivalent_Ybus\" finished in %s seconds:" % round((
                t_end - t_start), 2))
    return Ybus_eq


def adapt_impedance_params(Z, sign=1, adaption=1e-15):
    """
    In some extreme cases, the created admittance matrix of the
    zpbn network is singular. The routine is unsolvalbe with it.
    In response, an impedance adaption is created and added.
    """
    rft_pu = Z.real + sign * adaption
    xft_pu = Z.imag + sign * adaption
    return rft_pu, xft_pu


def _create_net_zpbn(net, boundary_buses, all_internal_buses, all_external_buses,
                     load_separate=False, sgen_separate=True, gen_separate=True,
                     show_computing_time=False, calc_volt_angles=True,
                     runpp_fct=_runpp_except_voltage_angles, **kwargs):
    """
    The function builds the zero power balance network with
    calculated impedance and voltage

    INPUT:
        **net** - pandapower network

        **boundary_buses** (list) - boundary buses

        **all_internal_buses** - all the internal buses

        **all_external_buses** - all the external buses

    OPTIONAL:
        **load_separate** (bool, False) - flag if all the loads
            are reserved integrally

        **sgen_separate** (bool, True) - flag if all the DER are
            reserved separately

        **gen_separate** (bool, True) - flag if all the gens are
            reserved separately

        **tolerance_mva** (float, 1e-3) - loadflow termination
            condition referring to P / Q mismatch of node power
            in MVA. The loalflow hier is to get the admittance
            matrix of the zpbn network

    OUTPUT:
        **net_zpbn** - zero power balance networks
    """

    net_internal, net_external = _get_internal_and_external_nets(
        net, boundary_buses, all_internal_buses, all_external_buses,
        show_computing_time, calc_volt_angles=calc_volt_angles, runpp_fct=runpp_fct, **kwargs)
    net_zpbn = net_external
    # --- remove buses without power flow results in net_eq
    drop_buses(net_zpbn, net_zpbn.res_bus.index[net_zpbn.res_bus.vm_pu.isnull()])

    Z, S, v, limits = calc_zpbn_parameters(net_zpbn, boundary_buses, all_external_buses)
    # --- remove the original load, sgen and gen in exteranl area,
    #     and creat new buses and impedance
    t_buses, g_buses = [], []
    sn_mva = net_zpbn.sn_mva
    for elm, separate in [("load", load_separate), ("sgen", sgen_separate), ("gen", gen_separate), ("ext_grid", False)]:
        # in Z columns only gen, load and sgens are considered, so we can leave out ext_grid
        net_zpbn[elm].drop(net_zpbn[elm].index[net_zpbn[elm].bus.isin(all_external_buses)], inplace=True)
        if elm == "ext_grid":
            continue

        if not np.isnan(Z[elm + "_ground"].values).all():
            if separate:
                Z = Z.drop([elm + "_integrated_total"], axis=1)

                # add buses
                idxs = Z.index[~np.isnan(Z[elm + "_ground"].values)]
                vn_kvs = net_zpbn.bus.vn_kv[Z.ext_bus.loc[idxs]]
                new_g_buses = create_buses(net_zpbn, len(idxs), vn_kvs, name=[
                    "%s_separate-ground %s" % (elm, str(Z.ext_bus.loc[i])) for i in idxs])
                new_t_buses = create_buses(net_zpbn, len(idxs), vn_kvs, name=[
                    "%s_separate-total %s" % (elm, str(Z.ext_bus.loc[i])) for i in idxs],
                                           max_vm_pu=limits.max_vm_pu.loc[idxs], min_vm_pu=limits.min_vm_pu.loc[idxs])

                # add impedances
                rft_pu_g, xft_pu_g = adapt_impedance_params(Z[elm + "_ground"].loc[idxs].values)
                max_idx = net_zpbn.impedance.index.max() if net_zpbn.impedance.shape[0] else 0
                new_imps_g = pd.DataFrame({
                    "from_bus": Z.ext_bus.loc[idxs].astype(np.int64).values, "to_bus": new_g_buses,
                    "rft_pu": rft_pu_g, "xft_pu": xft_pu_g,
                    "rtf_pu": rft_pu_g, "xtf_pu": xft_pu_g,
                    "gf_pu": 0, "bf_pu": 0, "gt_pu": 0, "bt_pu": 0},
                    index=range(max_idx + 1, max_idx + 1 + len(new_g_buses)))
                new_imps_g["name"] = "eq_impedance_ext_to_ground"
                new_imps_g["sn_mva"] = sn_mva
                new_imps_g["in_service"] = True

                rft_pu_t, xft_pu_t = adapt_impedance_params(Z[elm + "_separate_total"].loc[
                                                                idxs].values)
                new_imps_t = pd.DataFrame({
                    "from_bus": new_g_buses, "to_bus": new_t_buses,
                    "rft_pu": rft_pu_t, "xft_pu": xft_pu_t,
                    "rtf_pu": rft_pu_t, "xtf_pu": xft_pu_t,
                    "gf_pu": 0, "bf_pu": 0, "gt_pu": 0, "bt_pu": 0},
                    index=range(new_imps_g.index.max() + 1,
                                new_imps_g.index.max() + 1 + len(new_g_buses)))
                new_imps_t["name"] = "eq_impedance_ground_to_total"
                new_imps_t["sn_mva"] = sn_mva
                new_imps_t["in_service"] = True

                net_zpbn["impedance"] = pd.concat([net_zpbn["impedance"], new_imps_g, new_imps_t])
                g_buses += list(new_g_buses)
                t_buses += list(new_t_buses)
            else:
                Z = Z.drop([elm + "_separate_total"], axis=1)
                vn_kv = net_zpbn.bus.vn_kv[all_external_buses].values[0]
                new_g_bus = create_bus(net_zpbn, vn_kv, name=elm + "_integrated-ground ")
                i_all_integrated = []
                for i in Z.index[~np.isnan(Z[elm + "_ground"].values)]:
                    rft_pu, xft_pu = adapt_impedance_params(Z[elm + "_ground"][i])
                    create_impedance(net_zpbn, Z.ext_bus[i], new_g_bus, rft_pu, xft_pu,
                                     sn_mva, name="eq_impedance_ext_to_ground")
                    i_all_integrated.append(i)
                # in case of integrated, the tightest vm limits are assumed
                ext_buses = Z.ext_bus[~np.isnan(Z[elm + "_ground"])].values
                ext_buses_name = "/".join([str(eb) for eb in ext_buses])
                new_t_bus = create_bus(
                    net_zpbn, vn_kv, name=elm + "_integrated-total " + ext_buses_name,
                    max_vm_pu=limits.max_vm_pu.loc[i_all_integrated].min(),
                    min_vm_pu=limits.min_vm_pu.loc[i_all_integrated].max())
                rft_pu, xft_pu = adapt_impedance_params(Z[elm + "_integrated_total"][0])
                create_impedance(net_zpbn, new_g_bus, new_t_bus, rft_pu, xft_pu,
                                 sn_mva, name="eq_impedance_ground_to_total")
                g_buses += [new_g_bus.tolist()]
                t_buses += [new_t_bus.tolist()]
        else:
            Z.drop([elm + "_ground", elm + "_separate_total", elm + "_integrated_total"], axis=1,
                   inplace=True)

    # --- create load, sgen and gen
    elm_old = None
    max_load_idx = max(-1, net.load.index[~net.load.bus.isin(all_external_buses)].max() - len(net_zpbn.load))
    max_sgen_idx = max(-1, net.sgen.index[~net.sgen.bus.isin(all_external_buses)].max() - len(net_zpbn.sgen))
    max_gen_idx = max(-1, net.gen.index[~net.gen.bus.isin(all_external_buses)].max() - len(net_zpbn.gen))
    for i in t_buses:
        busstr = net_zpbn.bus.name[i].split(" ")[1]
        bus = int(busstr.split("/")[0])
        key = net_zpbn.bus.name[i].split("-")[0]
        elm = net_zpbn.bus.name[i].split("_")[0]
        idx = S.index[S.ext_bus == bus].values[0]
        P = S[key][idx].real * sn_mva
        Q = S[key][idx].imag * sn_mva
        Sn = S["sn_" + key][idx].real
        if elm == "load":
            elm_idx = create_load(net_zpbn, i, -float(P), -float(Q), name=key + "_rei_" + busstr,
                                  sn_mva=Sn, index=max_load_idx + len(net_zpbn.load) + 1)
        elif elm == "sgen":
            elm_idx = create_sgen(net_zpbn, i, float(P), float(Q), name=key + "_rei_" + busstr,
                                  sn_mva=Sn, index=max_sgen_idx + len(net_zpbn.sgen) + 1)
        elif elm == "gen":
            vm_pu = v[key + "_vm_total"][v.ext_bus == int(re.findall(r"\d+", busstr)[0])].values.real
            elm_idx = create_gen(net_zpbn, i, float(P), float(vm_pu.item()), name=key + "_rei_" + busstr,
                                 sn_mva=Sn, index=max_gen_idx + len(net_zpbn.gen) + 1)

        # ---- match other columns
        elm_org = net[elm]
        if elm_old is None or elm_old != elm:
            other_cols = set(elm_org.columns) - \
                         {"name", "bus", "p_mw", "q_mvar", "sn_mva", "in_service", "scaling"}
            other_cols_bool = set(
                net[elm][list(other_cols)].columns[net[elm][list(other_cols)].apply(is_bool_dtype)])
            other_cols -= other_cols_bool
            other_cols_number = set(
                net[elm][list(other_cols)].columns[
                    net[elm][list(other_cols)].apply(is_numeric_dtype)])
            other_cols -= other_cols_number
            other_cols_str = set()
            other_cols_none = set()
            other_cols_mixed = set()
            for c in other_cols.copy():
                value_types = net[elm][c][elm_org.bus.isin(all_external_buses)].apply(type).unique()
                if len(value_types) > 1:
                    other_cols_mixed |= {c}
                elif value_types[0] in (float, int):
                    other_cols_number |= {c}
                elif value_types[0] == bool:
                    other_cols_bool |= {c}
                elif value_types[0] == str:
                    other_cols_str |= {c}
                else:  # value_types[0] is None:
                    other_cols_none |= {c}
                other_cols -= {c}
            assert len(other_cols) == 0
        if "integrated" in key:
            if "voltLvl" in other_cols_number:
                net_zpbn[elm].loc[elm_idx, "voltLvl"] = \
                    net_zpbn.bus.voltLvl[boundary_buses].max()
                other_cols_number -= {"voltLvl"}
            net_zpbn[elm].loc[elm_idx, list(other_cols_number)] = \
                elm_org[list(other_cols_number)][elm_org.bus.isin(all_external_buses)].sum(axis=0)
            net_zpbn[elm].loc[elm_idx, list(other_cols_bool)] = elm_org[list(other_cols_bool)][
                                                                    elm_org.bus.isin(all_external_buses)].values.sum(
                axis=0) > 0
            all_str_values = list(zip(*elm_org[list(other_cols_str)] \
                                           [elm_org.bus.isin(all_external_buses)].values[::-1]))
            for asv, colid in zip(all_str_values, other_cols_str):
                if len(set(asv)) == 1:
                    net_zpbn[elm].loc[elm_idx, colid] = asv[0]
                else:
                    net_zpbn[elm].loc[elm_idx, colid] = "//".join(asv)
            net_zpbn[elm].loc[elm_idx, list(other_cols_none)] = None
            for ocm in other_cols_mixed:
                net_zpbn[elm][ocm] = net_zpbn[elm][ocm].astype("object")
            net_zpbn[elm].loc[elm_idx, list(other_cols_mixed)] = "mixed data type"
        else:
            if elm == "gen" and bus in net.ext_grid.bus.values and \
                    net.ext_grid.in_service[net.ext_grid.bus == bus].values[0]:
                net_zpbn[elm].loc[elm_idx, 'name'] = str(net.ext_grid.name[
                                                             net.ext_grid.bus == bus].values[0]) + "-" + \
                                                     net_zpbn[elm].name[elm_idx]
                ext_grid_cols = list(set(elm_org.columns) & set(net.ext_grid.columns) - \
                                     {"name", "bus", "p_mw", "sn_mva", "in_service", "scaling"})
                net_zpbn[elm].loc[elm_idx, ext_grid_cols] = net.ext_grid[ext_grid_cols][
                    net.ext_grid.bus == bus].values[0]
            else:
                names = elm_org.name[elm_org.bus == bus].values
                names = [str(n) for n in names]
                net_zpbn[elm].loc[elm_idx, "name"] = "//".join(names) + "-" + net_zpbn[elm].name[elm_idx]
                if len(names) > 1:
                    net_zpbn[elm].loc[elm_idx, list(other_cols_number)] = \
                        elm_org[list(other_cols_number)][elm_org.bus == bus].sum(axis=0)
                    if "voltLvl" in other_cols_number:
                        net_zpbn[elm].loc[elm_idx, "voltLvl"] = \
                            net_zpbn.bus.voltLvl[boundary_buses].max()
                    net_zpbn[elm].loc[elm_idx, list(other_cols_bool)] = \
                        elm_org[list(other_cols_bool)][elm_org.bus == bus].values.sum(axis=0) > 0

                    all_str_values = list(zip(*elm_org[list(other_cols_str)] \
                                                   [elm_org.bus == bus].values[::-1]))
                    for asv, colid in zip(all_str_values, other_cols_str):
                        if len(set(asv)) == 1:
                            net_zpbn[elm].loc[elm_idx, colid] = asv[0]
                        else:
                            net_zpbn[elm].loc[elm_idx, colid] = "//".join(asv)
                    net_zpbn[elm].loc[elm_idx, list(other_cols_none)] = None
                    for ocm in other_cols_mixed:
                        net_zpbn[elm][ocm] = net_zpbn[elm][ocm].astype("object")
                    net_zpbn[elm].loc[elm_idx, list(other_cols_mixed)] = "mixed data type"
                else:
                    net_zpbn[elm].loc[elm_idx, list(other_cols_bool | other_cols_number |
                                                    other_cols_str | other_cols_none)] = \
                        elm_org[list(other_cols_bool | other_cols_number |
                                     other_cols_str | other_cols_none)][
                            elm_org.bus == bus].values[0]
                    net_zpbn[elm].loc[elm_idx, list(other_cols)] = elm_org[list(other_cols)][
                        elm_org.bus == bus].values[0]
        elm_old = net_zpbn.bus.name[i].split("_")[0]

    # --- match poly_cost to new created elements
    for cost_elm in ["poly_cost", "pwl_cost"]:
        if len(net[cost_elm]):
            df = net_zpbn[cost_elm].copy()
            df.loc[(df.et == "ext_grid") &
                   (~df.bus.isin(boundary_buses)), 'et'] = "gen"
            df.loc[(df.et.isin(["storage", "dcline"]) &
                    (~df.bus.isin(boundary_buses))), 'et'] = "load"

            logger.debug("During the equivalencing, also in polt_cost, " +
                         "storages and dclines are treated as loads, and" +
                         "ext_grids are treated as gens ")

            for elm in ["load", "gen", "sgen"]:
                for idx in net_zpbn[elm].index:
                    if net_zpbn[elm].bus[idx] in boundary_buses:
                        continue
                    else:
                        pc_idx = df.index[df.et == elm]
                        if net_zpbn[elm].name.str.contains("integrated").any() and len(pc_idx):
                            logger.debug("Attention! After equivalencing, " + elm + "s are modeled as " +
                                         "an aggregated " + elm + ". The " + cost_elm + " data of the first " +
                                         "original " + elm + " is used as the " + cost_elm + " data of the " +
                                         "aggregated " + elm + ". It is NOT correct at present.")
                            df.element[pc_idx[0]] = net_zpbn[elm].index[net_zpbn[elm].name.str.contains(
                                "integrated", na=False)][0]
                            df = df.drop(pc_idx[1:])
                        elif len(pc_idx):
                            related_bus = int(str(net_zpbn[elm].name[idx]).split("_")[-1])
                            pc_idx = df.index[(df.bus == related_bus) &
                                              (df.et == elm)]
                            if len(pc_idx) > 1:
                                logger.debug("Attention! There are at least two " + elm + "s connected to a " +
                                             "common bus. The " + elm + "s with commen bus are modeled as an " +
                                             "aggreated " + elm + " during the equivalencing. " +
                                             "The " + cost_elm + " data of the first " + elm + " is used as the " +
                                             cost_elm + " data of the aggregated " + elm + ". " +
                                             "It is NOT correct at present.")
                                pc_idx = df.index[(df.bus == related_bus) &
                                                  (df.et == elm)]
                                df.element[pc_idx[0]] = idx
                                df = df.drop(pc_idx[1:])
                            elif len(pc_idx) == 1:
                                df.loc[pc_idx[0], 'element'] = idx
            net_zpbn[cost_elm] = df

    drop_and_edit_cost_functions(net_zpbn, [], False, True, False)
    # pp.runpp(net_zpbn)
    runpp_fct(net_zpbn, calculate_voltage_angles=calc_volt_angles,
              tolerance_mva=1e-3, max_iteration=100, **kwargs)
    return net_zpbn, net_internal, net_external


def _create_bus_lookups(net_zpbn, boundary_buses, all_internal_buses,
                        all_external_buses,
                        boundary_buses_inclusive_bswitch,
                        show_computing_time=False):
    """
    The function creates a bus lookup table according to the given zpbn network
    and the bus groups
    """
    t_start = time.perf_counter()
    build_ppc_and_Ybus(net_zpbn)
    ppc_branch = net_zpbn._ppc["branch"]
    pd2ppc_bus_lookup = net_zpbn._pd2ppc_lookups["bus"]

    # --- create bus lookup
    bus_lookup_pd = {"i_area_buses": [],
                     "b_area_buses": boundary_buses,
                     "t_area_buses": net_zpbn.bus.index[net_zpbn.bus.name.str.contains(
                         "-total", na=False)].tolist(),
                     "g_area_buses": net_zpbn.bus.index[net_zpbn.bus.name.str.contains(
                         "-ground", na=False)].tolist(),
                     "e_area_buses": all_external_buses,
                     "b_area_buses_no_switch": []}

    # --- create ppc bus lookup
    bus_lookup_ppc = bus_lookup_pd.copy()
    bus_lookup_ppc_origin = bus_lookup_pd.copy()
    for key in ["b", "i", "e", "g", "t"]:
        bus_lookup_ppc[key + "_area_buses"] = pd2ppc_bus_lookup[list(bus_lookup_pd[
                                                                         key + "_area_buses"])].tolist()
        origin_sequence = bus_lookup_ppc[key + "_area_buses"].copy()
        # remove repeated "neg." (e.g. bus-bus-switch) ppc buses
        bus_lookup_ppc[key + "_area_buses"] = sorted(set(bus_lookup_ppc[
                                                             key + "_area_buses"]), key=origin_sequence.index)
        if key == "b":
            # bus_lookup_pd["b_area_buses_no_switch"] = bus_lookup_pd[key+"_area_buses"].copy()
            # for i in range(len(bus_lookup_ppc[key+"_area_buses"])):
            #     if bus_lookup_ppc[key+"_area_buses"][i] != \
            #         pd2ppc_bus_lookup[bus_lookup_pd["b_area_buses_no_switch"][i]]:
            #             del bus_lookup_pd["b_area_buses_no_switch"][i]

            bus_lookup_pd["b_area_buses_no_switch"] = bus_lookup_pd[key + "_area_buses"].copy()
            val_ppc = []
            for val in bus_lookup_pd[key + "_area_buses"].copy():
                if pd2ppc_bus_lookup[val] in val_ppc:
                    bus_lookup_pd["b_area_buses_no_switch"].remove(val)
                else:
                    val_ppc.append(pd2ppc_bus_lookup[val])

        # remove ppc buses appearing in b_area as well as in i_area or e_area
        if key != "b":
            common_ppc_busese = set(bus_lookup_ppc["b_area_buses"]) & set(bus_lookup_ppc[
                                                                              key + "_area_buses"])
            if common_ppc_busese:
                bus_lookup_ppc[key + "_area_buses"].remove(list(common_ppc_busese))
        if key == "b" and len(bus_lookup_ppc["b_area_buses"]) != len(bus_lookup_pd["b_area_buses"]):
            logger.info("some boundary buses are connetected via switches")

    # --- identify "pos." (eg. bus-line-switch) ppc_buses
    Ybus_size = net_zpbn._ppc["internal"]["Ybus"]._shape[0]
    all_ppc_buses_lists = list(bus_lookup_ppc.values())
    all_ppc_buses = reduce(operator.concat, all_ppc_buses_lists)
    pos_aux_buses_ppc = set(np.arange(Ybus_size)) - set(all_ppc_buses)

    # --- identify the "pos." (eg. bus-line-switch) ppc buses belongs zu which bus group
    if pos_aux_buses_ppc:
        for br_idx in range(ppc_branch.shape[0]):
            f_bus_ppc = ppc_branch[br_idx, 0].real
            t_bus_ppc = ppc_branch[br_idx, 1].real
            if f_bus_ppc in pos_aux_buses_ppc:
                if t_bus_ppc in bus_lookup_ppc["e_area_buses"]:
                    bus_lookup_ppc["e_area_buses"].append(int(f_bus_ppc))
                    pos_aux_buses_ppc.remove(f_bus_ppc)
                elif t_bus_ppc in bus_lookup_ppc["i_area_buses"]:
                    bus_lookup_ppc["i_area_buses"].append(f_bus_ppc)
                    pos_aux_buses_ppc.remove(f_bus_ppc)
            elif t_bus_ppc in pos_aux_buses_ppc:
                if f_bus_ppc in bus_lookup_ppc["e_area_buses"]:
                    bus_lookup_ppc["e_area_buses"].append(int(t_bus_ppc))
                    pos_aux_buses_ppc.remove(t_bus_ppc)
                elif f_bus_ppc in bus_lookup_ppc["i_area_buses"]:
                    bus_lookup_ppc["i_area_buses"].append(int(t_bus_ppc))
                    pos_aux_buses_ppc.remove(t_bus_ppc)
        bus_lookup_ppc["e_area_buses"] += pos_aux_buses_ppc

    bus_lookups = ({"bus_lookup_pd": bus_lookup_pd,
                    "bus_lookup_ppc": bus_lookup_ppc,
                    "bus_lookup_ppc_origin": bus_lookup_ppc_origin,
                    "pos_aux_bus_ppc": pos_aux_buses_ppc,
                    "boundary_buses_inclusive_bswitch":
                        boundary_buses_inclusive_bswitch,
                    "origin_all_internal_buses": all_internal_buses})
    t_end = time.perf_counter()
    if show_computing_time:
        logger.info("\"create_bus_lookup\" finished in %s seconds:" % round((t_end - t_start), 2))
    return bus_lookups


def _get_internal_and_external_nets(net, boundary_buses, all_internal_buses,
                                    all_external_buses, show_computing_time=False,
                                    calc_volt_angles=True,
                                    runpp_fct=_runpp_except_voltage_angles, **kwargs):
    "This function identifies the internal area and the external area"
    t_start = time.perf_counter()
    if not all_internal_buses:
        net_internal = None
    else:
        net_internal = deepcopy(net)
        drop_measurements_and_controllers(net_internal, all_external_buses, True)
        drop_and_edit_cost_functions(net_internal,
                                     all_external_buses + boundary_buses,
                                     True, True)
        drop_buses(net_internal, all_external_buses)

    net_external = deepcopy(net)
    if "group" in net_external:
        net_external.group = net_external.group.drop(net_external.group.index)
    drop_and_edit_cost_functions(net_external, all_internal_buses,
                                 True, True)
    drop_measurements_and_controllers(net_external, net_external.bus.index.tolist())
    drop_buses(net_external, all_internal_buses)
    replace_motor_by_load(net_external, all_external_buses)
    #    add_ext_grids_to_boundaries(net_external, boundary_buses, runpp_fct=runpp_fct, **kwargs)
    #    runpp_fct(net_external, calculate_voltage_angles=calc_volt_angles, **kwargs)
    _integrate_power_elements_connected_with_switch_buses(net, net_external,
                                                          all_external_buses)  # for sgens, gens, and loads
    runpp_fct(net_external, calculate_voltage_angles=calc_volt_angles, **kwargs)
    t_end = time.perf_counter()
    if show_computing_time:
        logger.info("\"get_int_and_ext_nets\" " +
                    "finished in %s seconds:" % round((t_end - t_start), 2))

    return net_internal, net_external


def _calclate_equivalent_element_params(net_zpbn, Ybus_eq, bus_lookups,
                                        show_computing_time=False,
                                        max_allowed_impedance=1e8, **kwargs):
    """ This function calculates the equivalent parameters

    INPUT:
      **Ybus_eq** (array) - equivalent admittance matrix of the external area

      **bus_lookup** (dict) - bus lookup table

    OUTPUT:
      **shunt_params** - parameters of the equivalent shunts

      **impedance_params** - parameters of the equivalent impedances

    """
    t_start = time.perf_counter()
    # --- calculate impedance paramter
    bt_buses_ppc = list(bus_lookups["bus_lookup_ppc"]["b_area_buses"]) + \
                   list(bus_lookups["bus_lookup_ppc"]["t_area_buses"])
    bt_buses_pd = list(bus_lookups["bus_lookup_pd"]["b_area_buses_no_switch"]) + \
                  list(bus_lookups["bus_lookup_pd"]["t_area_buses"])
    nb_bt_buses_ppc = len(bt_buses_ppc)

    shunt_params = pd.DataFrame(columns=["bus_pd", "bus_ppc", "parameter"])
    shunt_params["bus_ppc"] = bt_buses_ppc
    shunt_params["bus_pd"] = bt_buses_pd
    shunt_params["parameter"] = Ybus_eq.sum(axis=1)[-nb_bt_buses_ppc:]
    shunt_params["local_voltage"] = net_zpbn.res_bus.vm_pu[bt_buses_pd].values

    # --- calculate impedance paramter
    params = Ybus_eq[-nb_bt_buses_ppc:, -nb_bt_buses_ppc:]
    tri_upper = np.triu(params, k=1)
    non_zero = np.abs(tri_upper) > 1 / max_allowed_impedance
    rows = (np.arange(params.shape[0]).reshape(-1, 1) * np.ones(params.shape)).astype(np.int64)[non_zero]
    cols = (np.arange(params.shape[1]) * np.ones(params.shape)).astype(np.int64)[non_zero]

    impedance_params = pd.DataFrame(columns=impedance_columns, index=range(len(rows)), data=0)
    impedance_params["from_bus"] = np.array(bt_buses_pd)[rows]
    impedance_params["to_bus"] = np.array(bt_buses_pd)[cols]
    impedance_params["rft_pu"] = (-1 / params[rows, cols]).real
    impedance_params["xft_pu"] = (-1 / params[rows, cols]).imag

    non_zero_cr = np.abs(params[cols, rows]) > 1 / max_allowed_impedance
    impedance_params["rtf_pu"] = 1e5
    impedance_params["xtf_pu"] = 1e5
    impedance_params.loc[non_zero_cr, "rtf_pu"] = (-1 / params[cols, rows]).real[non_zero_cr]
    impedance_params.loc[non_zero_cr, "xtf_pu"] = (-1 / params[cols, rows]).imag[non_zero_cr]

    t_end = time.perf_counter()
    if show_computing_time:
        logger.info("\"calclate_equivalent_element_params\" finished in %s seconds:" %
                    round((t_end - t_start), 2))
    return shunt_params, impedance_params


def _replace_ext_area_by_impedances_and_shunts(
        net_eq, bus_lookups, impedance_params, shunt_params, net_internal,
        return_internal, show_computing_time=False, calc_volt_angles=True, imp_threshold=1e-8,
        runpp_fct=_runpp_except_voltage_angles, **kwargs):
    """
    This function implements the parameters of the equivalent shunts and equivalent impedance
    """
    # --- drop all external elements
    eg_buses_pd = bus_lookups["bus_lookup_pd"]["e_area_buses"] + \
                  bus_lookups["bus_lookup_pd"]["g_area_buses"]
    drop_buses(net_eq, eg_buses_pd)

    try:
        runpp_fct(net_eq, calculate_voltage_angles=calc_volt_angles,
                  tolerance_mva=1e-6, max_iteration=100, **kwargs)
    except LoadflowNotConverged:
        logger.error("The power flow did not converge.")

    # --- drop all branch elements except switches between boundary buses
    drop_internal_branch_elements(net_eq, bus_lookups["boundary_buses_inclusive_bswitch"])
    # --- drop shunt elements attached to boundary buses
    traget_shunt_idx = net_eq.shunt.index[net_eq.shunt.bus.isin(bus_lookups[
                                                                    "boundary_buses_inclusive_bswitch"])]
    net_eq.shunt = net_eq.shunt.drop(traget_shunt_idx)

    # --- create impedance
    not_very_low_imp = (impedance_params.rft_pu.abs() > imp_threshold) | (
            impedance_params.xft_pu.abs() > imp_threshold) | (
                               impedance_params.rtf_pu.abs() > imp_threshold) | (
                               impedance_params.xtf_pu.abs() > imp_threshold) | (
                               impedance_params.from_bus.isin(set(net_eq.gen.bus) | set(net_eq.ext_grid.bus)) &
                               impedance_params.to_bus.isin(set(net_eq.gen.bus) | set(net_eq.ext_grid.bus)))
    new_imps = impedance_params.loc[not_very_low_imp, impedance_columns]
    max_idx = net_eq.impedance.index.max() if net_eq.impedance.shape[0] else 0
    new_imps.index = range(max_idx + 1, max_idx + 1 + sum(not_very_low_imp))
    new_imps["name"] = "eq_impedance"
    new_imps["sn_mva"] = net_eq.sn_mva
    new_imps["in_service"] = True
    net_eq["impedance"] = pd.concat([net_eq["impedance"], new_imps])

    # --- create switches instead of very low impedances
    new_sws = impedance_params[["from_bus", "to_bus"]].loc[~not_very_low_imp].astype(np.int64)
    new_sws = new_sws.rename(columns={"from_bus": "bus", "to_bus": "element"})
    max_idx = net_eq.switch.index.max() if net_eq.switch.shape[0] else 0
    new_sws.index = range(max_idx + 1, max_idx + 1 + sum(~not_very_low_imp))
    new_sws["et"] = "b"
    new_sws["name"] = "eq_switch"
    new_sws["closed"] = True
    new_sws["z_ohm"] = 0
    net_eq["switch"] = pd.concat([net_eq["switch"], new_sws])
    # If some buses are connected through switches, their shunts are connected in parallel
    # to same bus. The shunt parameters needs to be adapted. TODO
    if not not_very_low_imp.all():
        fb = impedance_params.from_bus[~not_very_low_imp].values.tolist()
        tb = impedance_params.to_bus[~not_very_low_imp].values.tolist()
        # fb_values = shunt_params.parameter[shunt_params.bus_pd.isin(fb)].values
        # tb_values = shunt_params.parameter[shunt_params.bus_pd.isin(tb)].values
        # adapted_params = fb_values * tb_values / (tb_values + fb_values)
        # shunt_params.parameter[shunt_params.bus_pd.isin(tb)] = adapted_params
        shunt_params.drop(shunt_params.index[shunt_params.bus_pd.isin(fb)], inplace=True)
        shunt_params.drop(shunt_params.index[shunt_params.bus_pd.isin(tb)], inplace=True)

    # --- create shunts
    max_idx = net_eq.shunt.index.max() if net_eq.shunt.shape[0] else 0
    shunt_buses = shunt_params.bus_pd.values.astype(np.int64)
    new_shunts = pd.DataFrame({"bus": shunt_buses,
                               "q_mvar": -shunt_params.parameter.values.imag * net_eq.sn_mva,
                               "p_mw": shunt_params.parameter.values.real * net_eq.sn_mva
                               }, index=range(max_idx + 1, max_idx + 1 + shunt_params.shape[0]))
    new_shunts["name"] = "eq_shunt"
    isin_sh = new_shunts.bus.isin(net_eq.bus.index)
    new_shunts.loc[isin_sh, "vn_kv"] = net_eq.bus.vn_kv.loc[new_shunts.bus.loc[isin_sh]].values
    new_shunts["step"] = 1
    new_shunts["max_step"] = 1
    new_shunts["in_service"] = True
    if "step_dependency_table" in net_eq.shunt:
        new_shunts["step_dependency_table"] = False
    net_eq["shunt"] = pd.concat([net_eq["shunt"], new_shunts])
    if n_disconnected_new_eq_shunts := sum(~isin_sh):
        msg = f"{n_disconnected_new_eq_shunts=}, missing buses: {new_shunts.bus.loc[~isin_sh]}"
        raise ValueError(msg)

    runpp_fct(net_eq, calculate_voltage_angles=calc_volt_angles,
              tolerance_mva=1e-6, max_iteration=100, **kwargs)


def _integrate_power_elements_connected_with_switch_buses(net, net_external, all_external_buses):
    _, bus_dict = get_connected_switch_buses_groups(net_external, all_external_buses)
    active_sgen = net.sgen[(net.sgen.in_service) & ~((net["sgen"].p_mw == 0) & (net["sgen"].q_mvar == 0))]
    active_load = net.load[(net.load.in_service) & ~((net["load"].p_mw == 0) & (net["load"].q_mvar == 0))]
    active_gen = net.gen[net.gen.in_service]

    for elm in ["sgen", "load", "gen"]:
        for bd in bus_dict:
            if elm == "sgen":
                connected_elms = active_sgen.index[active_sgen.bus.isin(bd)]
            elif elm == "load":
                connected_elms = active_load.index[active_load.bus.isin(bd)]
            elif elm == "gen":
                connected_elms = active_gen.index[active_gen.bus.isin(bd)]
            else:
                raise ValueError(f"{elm=} not supported in "
                                 "_integrate_power_elements_connected_with_switch_buses()")
            if len(connected_elms) <= 1:
                continue
            else:  # There ars some "external" elements connected with bus-bus switches.
                # They will be aggregated.
                elm1 = connected_elms[0]
                net[elm].loc[connected_elms, 'bus'] = net[elm].bus[elm1]
                net_external[elm].loc[connected_elms, 'bus'] = net_external[elm].bus[elm1]



# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from numpy import zeros, array, int64
from pandapower.pypower.idx_cost import MODEL, NCOST, COST, PW_LINEAR, POLYNOMIAL
from pandapower.pypower.idx_gen import PMIN, PMAX

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def _make_objective(ppci, net):
    is_quadratic, q_costs = _init_gencost(ppci, net)
    if len(net.pwl_cost):
        ppci["gencost"][:, MODEL] = PW_LINEAR
        ppci["gencost"][:, NCOST] = 2
        ppci["gencost"][:, COST + 2] = 1

        _fill_gencost_pwl(ppci, net)
        if is_quadratic:
            raise ValueError("Piecewise linear costs can not be mixed with quadratic costs")
        elif len(net.poly_cost):
            _add_linear_costs_as_pwl_cost(ppci, net)
    elif len(net.poly_cost):
        ppci["gencost"][:, MODEL] = POLYNOMIAL
        _fill_gencost_poly(ppci, net, is_quadratic, q_costs)
    else:
        logger.warning("no costs are given - overall generated power is minimized")
        ppci["gencost"][:, MODEL] = POLYNOMIAL
        ppci["gencost"][:, NCOST] = 2
        ppci["gencost"][:, COST] = 1
    return ppci


def _get_gen_index(net, et, element):
    if et == "dcline":
        dc_idx = net.dcline.index.get_loc(element)
        element = len(net.gen.index) - 2*len(net.dcline) + dc_idx*2 + 1
        et = "gen"
    lookup = "%s_controllable" % et if et in ["load", "sgen", "storage"] else et
    try:
        return int(net._pd2ppc_lookups[lookup][int(element)])
    except:
        return None


def _map_costs_to_gen(net, cost):
    gens = array([_get_gen_index(net, et, element)
                  for et, element in zip(cost.et.values, cost.element.values)])
    cost_is = array([gen is not None for gen in gens])
    cost = cost[cost_is]
    gens = gens[cost_is].astype(int64)
    signs = array([-1 if element in ["load", "storage", "dcline"] else 1 for element in cost.et])
    return gens, cost, signs


def _init_gencost(ppci, net):
    is_quadratic = net.poly_cost[["cp2_eur_per_mw2", "cq2_eur_per_mvar2"]].values.any()
    q_costs = net.poly_cost[["cq1_eur_per_mvar", "cq2_eur_per_mvar2"]].values.any() or \
        "q" in net.pwl_cost.power_type.values
    rows = len(ppci["gen"])*2 if q_costs else len(ppci["gen"])
    if len(net.pwl_cost):
        nr_points = {len(p) for p in net.pwl_cost.points.values}
        points = max(nr_points)
        if is_quadratic:
            raise ValueError("Quadratic costs can be mixed with piecewise linear costs")
        columns = COST + (max(points, 2) + 1)*2
    else:
        columns = COST + 3 if is_quadratic else COST + 2
    ppci["gencost"] = zeros((rows, columns), dtype=float)
    return is_quadratic, q_costs


def _fill_gencost_poly(ppci, net, is_quadratic, q_costs):
    gens, cost, signs = _map_costs_to_gen(net, net.poly_cost)
    c0 = cost["cp0_eur"].values
    c1 = cost["cp1_eur_per_mw"].values
    signs = array([-1 if element in ["load", "storage", "dcline"] else 1 for element in cost.et])
    if is_quadratic:
        c2 = cost["cp2_eur_per_mw2"]
        ppci["gencost"][gens, NCOST] = 3
        ppci["gencost"][gens, COST] = c2 * signs
        ppci["gencost"][gens, COST + 1] = c1 * signs
        ppci["gencost"][gens, COST + 2] = c0 * signs
    else:
        ppci["gencost"][gens, NCOST] = 2
        ppci["gencost"][gens, COST] = c1 * signs
        ppci["gencost"][gens, COST + 1] = c0 * signs
    if q_costs:
        gens_q = gens + len(ppci["gen"])
        c0 = cost["cq0_eur"].values
        c1 = cost["cq1_eur_per_mvar"].values
        signs = array([-1 if element in ["load", "storage"] else 1 for element in cost.et])
        if is_quadratic:
            c2 = cost["cq2_eur_per_mvar2"]
            ppci["gencost"][gens_q, NCOST] = 3
            ppci["gencost"][gens_q, COST] = c2 * signs
            ppci["gencost"][gens_q, COST + 1] = c1 * signs
            ppci["gencost"][gens_q, COST + 2] = c0 * signs
        else:
            ppci["gencost"][gens_q, NCOST] = 2
            ppci["gencost"][gens_q, COST] = c1 * signs
            ppci["gencost"][gens_q, COST + 1] = c0 * signs


def _fill_gencost_pwl(ppci, net):
    for power_mode, cost in net.pwl_cost.groupby("power_type"):
        gens, cost, signs = _map_costs_to_gen(net, cost)
        if power_mode == "q":
            gens += len(ppci["gen"])
        for gen, points, sign in zip(gens, cost.points.values, signs):
            costs = costs_from_areas(points, sign)
            ppci["gencost"][gen, COST:COST+len(costs)] = costs
            ppci["gencost"][gen, NCOST] = len(costs) / 2


def costs_from_areas(points, sign):
    costs = []
    c0 = 0
    last_upper = None
    for lower, upper, slope in points:
        if last_upper is None:
            costs.append(lower)
            c = c0 + lower * slope * sign
            c0 = c
            costs.append(c)
        if last_upper is not None and last_upper != lower:
            raise ValueError("Non-consecutive cost function areas")
        last_upper = upper
        costs.append(upper)
        c = c0 + (upper - lower) * slope * sign
        c0 = c
        costs.append(c)
    return costs


def _add_linear_costs_as_pwl_cost(ppci, net):
    gens, cost, signs = _map_costs_to_gen(net, net.poly_cost)
    ppci["gencost"][gens, NCOST] = 2
    pmin = ppci["gen"][gens, PMIN]
    pmax = ppci["gen"][gens, PMAX]
    ppci["gencost"][gens, COST] = pmin
    ppci["gencost"][gens, COST + 1] = pmin * cost.cp1_eur_per_mw.values * signs
    ppci["gencost"][gens, COST + 2] = pmax
    ppci["gencost"][gens, COST + 3] = pmax * cost.cp1_eur_per_mw.values * signs



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import numpy as np
import pandas as pd
import warnings
from packaging.version import Version

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)
try:
    import lightsim2grid
    v = Version(lightsim2grid.__version__)
    if v < Version("0.9.0"):
        logger.warning("Only lightsim2grid version 0.9.0 or newer is supported - please update ligtsim2grid")
        raise ImportError

    from lightsim2grid.gridmodel.from_pandapower import init as init_ls2g
    from lightsim2grid.contingencyAnalysis import ContingencyAnalysisCPP
    from lightsim2grid_cpp import SolverType

    lightsim2grid_installed = True
except ImportError:
    lightsim2grid_installed = False

try:
    from lightsim2grid_cpp import KLUSolver, KLUSolverSingleSlack

    KLU_solver_available = True
except ImportError:
    KLU_solver_available = False

from pandapower.run import runpp


def run_contingency(net, nminus1_cases, pf_options=None, pf_options_nminus1=None, write_to_net=True,
                    contingency_evaluation_function=runpp, **kwargs):
    """
    Obtain either loading (N-0) or max. loading (N-0 and all N-1 cases), and min/max bus voltage magnitude.
    The variable "temperature_degree_celsius" can be used in addition to "loading_percent" to obtain max. temperature.
    In the returned dictionary, the variable "loading_percent" represents the loading in N-0 case,
    "max_loading_percent" and "min_loading_percent" represent highest and lowest observed loading_percent among all
    calculated N-1 cases. The same convention applies to "temperature_degree_celsius" when applicable.
    This function can be passed through to pandapower.timeseries.run_timeseries as the run_control_fct argument.

    INPUT
    ----------
    **net** - pandapowerNet
    **nminus1_cases** - dict
        describes all N-1 cases, e.g. {"line": {"index": [1, 2, 3]}, "trafo": {"index": [0]}, "trafo3w": {"index": [1]}}
    **pf_options** - dict
        options for power flow calculation in N-0 case
    **pf_options_nminus1** - dict
        options for power flow calculation in N-1 cases
    **write_to_net** - bool
        whether to write the results of contingency analysis to net (in "res_" tables). The results will be written for
        the following additional variables: table res_bus with columns "max_vm_pu", "min_vm_pu",
        tables res_line, res_trafo, res_trafo3w with columns "max_loading_percent", "min_loading_percent",
        "causes_overloading", "cause_element", "cause_index", table res_line with columns
        "max_temperature_degree_celsius", "min_temperature_degree_celsius" (if "tdpf" set to True)
        "causes_overloading": does this element, when defining the N-1 case, cause overloading of other elements? the
        overloading is defined by net.line["max_loading_percent_nminus1"] (if set) or net.line["max_loading_percent"]
        "cause_element": element ("line", "trafo", "trafo3w") that causes max. loading of this element
        "cause_index": index of the element ("line", "trafo", "trafo3w") that causes max. loading of this element
    **contingency_evaluation_function** - func
        function to use for power flow calculation, default pp.runpp

    OUTPUT
    -------
    **contingency_results** - dict
        dict of arrays per element for index, min/max result
    """
    # set up the dict for results and relevant variables
    # ".get" in case the options have been set in pp.set_user_pf_options:
    raise_errors = kwargs.get("raise_errors", False)
    if "recycle" in kwargs:
        kwargs["recycle"] = False  # so that we can be sure it doesn't happen
    if pf_options is None:
        pf_options = net.user_pf_options.get("pf_options", net.user_pf_options)
    if pf_options_nminus1 is None:
        pf_options_nminus1 = net.user_pf_options.get("pf_options_nminus1", net.user_pf_options)
    if kwargs is not None:
        # avoid duplicate passing keys to contingency_evaluation_function()
        pf_options = {key: val for key, val in pf_options.items() if key not in kwargs.keys()}
        pf_options_nminus1 = {key: val for key, val in pf_options_nminus1.items() if key not in
                              kwargs.keys()}

    contingency_results = {element: {"index": net[element].index.values}
                           for element in ("bus", "line", "trafo", "trafo3w") if len(net[element]) > 0}
    for element in contingency_results.keys():
        if element == "bus":
            continue
        contingency_results[element].update(
            {"causes_overloading": np.zeros_like(net[element].index.values, dtype=bool),
             "cause_element": np.empty_like(net[element].index.values, dtype=object),
             "cause_index": np.empty_like(net[element].index.values, dtype=np.int64)})
    result_variables = {**{"bus": ["vm_pu"]},
                        **{key: ["loading_percent"] for key in ("line", "trafo", "trafo3w") if len(net[key]) > 0}}
    if len(net.line) > 0 and (net.get("_options", {}).get("tdpf", False) or
                              pf_options.get("tdpf", False) or pf_options_nminus1.get("tdpf", False)):
        result_variables["line"].append("temperature_degree_celsius")

    # for n-1
    for element, val in nminus1_cases.items():
        for i in val["index"]:
            if not net[element].at[i, "in_service"]:
                continue
            net[element].at[i, 'in_service'] = False
            try:
                contingency_evaluation_function(net, **pf_options_nminus1, **kwargs)
                _update_contingency_results(net, contingency_results, result_variables, nminus1=True,
                                            cause_element=element, cause_index=i)
            except Exception as err:
                logger.error(f"{element} {i} causes {err}")
                if raise_errors:
                    raise err
            finally:
                net[element].at[i, 'in_service'] = True

    # for n-0
    contingency_evaluation_function(net, **pf_options, **kwargs)
    _update_contingency_results(net, contingency_results, result_variables, nminus1=False)

    if write_to_net:
        for element, element_results in contingency_results.items():
            index = element_results["index"]
            for var, val in element_results.items():
                if var == "index" or var in net[f"res_{element}"].columns.values:
                    continue
                net[f"res_{element}"].loc[index, var] = val

    return contingency_results


def run_contingency_ls2g(net, nminus1_cases, contingency_evaluation_function=runpp, **kwargs):
    """
    Execute contingency analysis using the lightsim2grid library. This works much faster than using
    pandapower.
    This function can be passed through to pandapower.timeseries.run_timeseries as the
    run_control_fct argument.

    **Limitation:** the results for branch flows are valid only for the "from_bus" of lines and
    "hv_bus" of transformers. This can lead to a small difference to the results using pandapower.
    The results are written in pandapower results tables. Make sure that the N-1 cases do not lead
    to isolated grid, otherwise results with pandapower and this function will
    be different. Reason: pandapower selects a different gen as slack if the grid becomes isolated,
    but lightsim2grid would simply return nan as results for such a contingency situation.

    **WARNING:** continuous bus indices, 0-start, are required!

    The results will written for the
    following additional variables: table res_bus with columns "max_vm_pu", "min_vm_pu",
    tables res_line, res_trafo, res_trafo3w with columns "max_loading_percent", "min_loading_percent",
    "causes_overloading", "cause_element", "cause_index", table res_line with columns "max_temperature_degree_celsius",
    "min_temperature_degree_celsius" (if "tdpf" set to True)
    "causes_overloading": does this element, when defining the N-1 case, cause overloading of other elements? the
    overloading is defined by net.line["max_loading_percent_nminus1"] (if set) or net.line["max_loading_percent"]
    "cause_element": element ("line", "trafo", "trafo3w") that causes max. loading of this element
    "cause_index": index of the element ("line", "trafo", "trafo3w") that causes max. loading of this element
    "congestion_caused_mva": overall congestion in the grid in MVA during the N-1 case due to the failure of the element

    INPUT
    ----------

    **net** - pandapowerNet
    **nminus1_cases** - dict
        describes all N-1 cases, e.g. {"line": {"index": [1, 2, 3]}, "trafo": {"index": [0]}}
        Note: trafo3w is not supported
    **contingency_evaluation_function** - func
        function to use for power flow calculation, default pp.runpp (but only relevant for N-0 case)
    """
    if not lightsim2grid_installed:
        raise UserWarning("lightsim2grid package not installed. "
                          "Install lightsim2grid e.g. by running 'pip install lightsim2grid' in command prompt.")
    # check for continuous bus index starting with 0:
    n_bus = len(net.bus)
    last_bus = net.bus.index[-1]
    if net.bus.index[0] != 0 or last_bus != n_bus - 1 or sum(net.bus.index) != last_bus * n_bus / 2:
        raise UserWarning("bus index must be continuous and start with 0 "
                          "(use pandapower.create_continuous_bus_index)")
    contingency_evaluation_function(net, **kwargs)

    tps_flag = False
    tps, tps_tap_pos, tps_shift_degree = None, None, None
    if "tap_phase_shifter" in net.trafo.columns:
        if np.any(net.trafo.tap_phase_shifter):
            tps_flag = True
            tps, tps_tap_pos, tps_shift_degree = _convert_trafo_phase_shifter(
                net, "trafo", "tap_phase_shifter")

    tct2w_flag = False
    tct2w, tct2w_tap_pos, tct2w_shift_degree = None, None, None
    tct3w_flag = False
    tct3w, tct3w_tap_pos, tct3w_shift_degree = None, None, None
    if "tap_changer_type" in net.trafo.columns:
        if np.any(net.trafo.tap_changer_type == "Ideal"):
            tct2w_flag = True
            tct2w, tct2w_tap_pos, tct2w_shift_degree = _convert_trafo_phase_shifter(
                net, "trafo", "tap_changer_type")
    if "tap_changer_type" in net.trafo3w.columns:
        if np.any(net.trafo3w.tap_changer_type == "Ideal"):
            tct3w_flag = True
            tct3w, tct3w_tap_pos, tct3w_shift_degree = _convert_trafo_phase_shifter(
                    net, "trafo3w", "tap_changer_type")

    # setting "slack" back-and-forth is due to the difference in interpretation of generators as "distributed slack"
    if net._options.get("distributed_slack", False):
        slack_backup = net.gen.slack.copy()
        net.gen.loc[net.gen.slack_weight != 0, 'slack'] = True
        msg = "LightSim cannot handle multiple slack bus at the moment. Only the first " \
              "slack bus of pandapower will be used."
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", msg)
            lightsim_grid_model = init_ls2g(net)
        net.gen['slack'] = slack_backup
        solver_type = SolverType.KLU if KLU_solver_available else SolverType.SparseLU
    else:
        lightsim_grid_model = init_ls2g(net)
        solver_type = SolverType.KLUSingleSlack if KLU_solver_available else SolverType.SparseLUSingleSlack

    if tps_flag:
        net.trafo.tap_phase_shifter = tps
        net.trafo.tap_pos = tps_tap_pos
        net.trafo.shift_degree = tps_shift_degree
    if tct2w_flag:
        net.trafo.tap_changer_type = tct2w
        net.trafo.tap_pos = tct2w_tap_pos
        net.trafo.shift_degree = tct2w_shift_degree
    if tct3w_flag:
        net.trafo3w.tap_changer_type = tct3w
        net.trafo3w.tap_pos = tct3w_tap_pos
        net.trafo3w.shift_degree = tct3w_shift_degree

    n_lines = len(net.line)
    n_lines_cases = len(nminus1_cases.get("line", {}).get("index", []))
    n_trafos = len(net.trafo)
    n_trafos_cases = len(nminus1_cases.get("trafo", {}).get("index", []))

    # todo: add option for DC power flow
    s = ContingencyAnalysisCPP(lightsim_grid_model)
    s.change_solver(solver_type)

    map_index = {}
    for element, values in nminus1_cases.items():
        index = np.array(_get_iloc_index(net[element].index.values, values["index"]), dtype=np.int64)
        map_index[element] = index.copy()  # copy() because += n_lines happens later
        if element == "trafo3w":
            raise NotImplementedError("trafo3w not implemented for lightsim2grid contingency analysis")
        elif element == "trafo":
            index += n_lines
        s.add_multiple_n1(index)

    # s.add_multiple_n1(net.line.index.values.astype(int))
    v_init = net._ppc["internal"]["V"]
    s.compute(v_init, net._options["max_iteration"], net._options["tolerance_mva"])
    v_res = s.get_voltages()
    s.compute_flows()
    kamps_all = s.get_flows()

    vm_pu = np.abs(v_res)
    net.res_bus["max_vm_pu"] = np.nanmax(vm_pu, axis=0)
    net.res_bus["min_vm_pu"] = np.nanmin(vm_pu, axis=0)

    max_i_ka_limit_all = np.r_[
        net.line.max_i_ka.values, net.trafo.sn_mva.values / (net.trafo.vn_hv_kv.values * np.sqrt(3))]
    max_loading_limit_all = np.r_[net.line["max_loading_percent_nminus1"]
    if "max_loading_percent_nminus1" in net.line.columns
    else net.line["max_loading_percent"] if n_lines > 0 else [],
    net.trafo["max_loading_percent_nminus1"]
    if "max_loading_percent_nminus1" in net.trafo.columns
    else net.trafo["max_loading_percent"] if n_trafos > 0 else []]
    voltage_all = np.r_[net.bus.loc[net.line.from_bus.values, "vn_kv"].values if n_lines > 0 else [],
    net.trafo.vn_hv_kv if n_trafos > 0 else []]
    flows_all_mva = np.nan_to_num(kamps_all * voltage_all * np.sqrt(3))
    flows_limit_all = np.nan_to_num(max_loading_limit_all / 100 * max_i_ka_limit_all * voltage_all * np.sqrt(3))

    big_number = 1e6
    for element in ("line", "trafo"):
        if len(net[element]) == 0:
            continue
        if element == "line":
            kamps_element = kamps_all[:, 0:n_lines]
            kamps_element_cause = kamps_all[0:n_lines_cases, :]
            flows_element_mva = flows_all_mva[0:n_lines_cases, :]
            max_i_ka_limit = max_i_ka_limit_all[0:n_lines]
        else:
            kamps_element = kamps_all[:, n_lines:n_lines + n_trafos]
            kamps_element_cause = kamps_all[n_lines_cases:n_lines_cases + n_trafos_cases, :]
            flows_element_mva = flows_all_mva[n_lines_cases:n_lines_cases + n_trafos_cases, :]
            max_i_ka_limit = max_i_ka_limit_all[n_lines:n_lines + n_trafos]
        # max_i_ka = np.nanmax(kamps_line, where=~np.isnan(kamps_line), axis=0, initial=0)
        cause_index = np.nanargmax(kamps_element, axis=0)
        cause_element = np.where(cause_index < n_lines_cases, "line", "trafo")
        max_i_ka = np.nanmax(kamps_element, axis=0)
        net[f"res_{element}"]["max_loading_percent"] = max_i_ka / max_i_ka_limit * 100
        min_i_ka = np.nanmin(kamps_element, axis=0, where=kamps_element != 0,
                             initial=big_number)  # this is quite daring tbh
        min_i_ka[min_i_ka == big_number] = 0
        # min_i_ka = np.nanmin(kamps_line, axis=0)
        net[f"res_{element}"]["min_loading_percent"] = min_i_ka / max_i_ka_limit * 100
        causes_overloading = np.any(kamps_element_cause > max_loading_limit_all * max_i_ka_limit_all / 100, axis=1)
        net[f"res_{element}"]["causes_overloading"] = False
        if element in nminus1_cases:
            # order of n-1 cases is always sorted, so "vertical" sorting is different than "horizontal"
            net[f"res_{element}"].loc[net[element].index.values[
                np.sort(map_index[element])], "causes_overloading"] = causes_overloading
        cause_mask = cause_element == "line"
        if "line" in map_index:
            cause_index[cause_mask] = net.line.index.values[np.sort(map_index["line"])[cause_index[cause_mask]]]
        if "trafo" in map_index:
            cause_index[~cause_mask] = net.trafo.index.values[
                np.sort(map_index["trafo"])[cause_index[~cause_mask] - n_lines_cases]]
        net[f"res_{element}"]["cause_index"] = cause_index
        net[f"res_{element}"]["cause_element"] = cause_element

        congestion_mva = flows_element_mva - flows_limit_all
        congestion_mva = np.where(congestion_mva < 0, 0, congestion_mva)
        congestion_caused = congestion_mva.sum(axis=1)
        if element in nminus1_cases:
            # order of n-1 cases is always sorted, so "vertical" sorting is different than "horizontal"
            net[f"res_{element}"].loc[net[element].index.values[
                np.sort(map_index[element])], "congestion_caused_mva"] = congestion_caused


def _convert_trafo_phase_shifter(net, trafotable="trafo", tap_param="tap_changer_type"):
    """
    Moves tap_step_degree values for ideal phase shifters to the shift_degree values.
    Works both for old implementation (tap_phase_shifter) and for new implementation (tap_changer_type).
    """
    tap_changer = net[trafotable][tap_param].values.copy()
    shift_degree = net[trafotable].shift_degree.values.copy()

    tap_pos = net[trafotable].tap_pos.values
    tap_neutral = net[trafotable].tap_neutral.values
    tap_diff = tap_pos - tap_neutral
    tap_step_degree = net[trafotable].tap_step_degree.values.copy()

    if tap_param == "tap_changer_type":
        net.trafo.loc[tap_changer == "Ideal", 'shift_degree'] += (tap_diff[tap_changer == "Ideal"] *
                                                                  tap_step_degree[tap_changer == "Ideal"])
        net[trafotable][tap_param] = None
    elif tap_param == "tap_phase_shifter":
        net.trafo.loc[tap_changer, 'shift_degree'] += tap_diff[tap_changer] * tap_step_degree[tap_changer]
        net[trafotable][tap_param] = False
    else:
        raise ValueError("tap_param argument not recognized")

    net[trafotable]["tap_pos"] = 0

    return tap_changer, tap_pos, shift_degree


def _update_contingency_results(net, contingency_results, result_variables, nminus1, cause_element=None,
                                cause_index=None):
    for element, vars in result_variables.items():
        for var in vars:
            val = net[f"res_{element}"][var].values
            if nminus1:
                if var == "loading_percent":
                    s = 'max_loading_percent_nminus1' \
                        if 'max_loading_percent_nminus1' in net[element].columns \
                        else 'max_loading_percent'
                    # this part with cause_mask and max_mask is not very efficient nor orderly
                    loading_limit = net[element].loc[contingency_results[element]["index"], s].values
                    cause_mask = val > loading_limit
                    if np.any(cause_mask):
                        contingency_results[cause_element]["causes_overloading"][
                            contingency_results[cause_element]["index"] == cause_index] = True
                    max_mask = val > contingency_results[element].get("max_loading_percent", np.full_like(val, -1))
                    if np.any(max_mask):
                        contingency_results[element]["cause_index"][max_mask] = cause_index
                        contingency_results[element]["cause_element"][max_mask] = cause_element
                for func, min_max in ((np.fmax, "max"), (np.fmin, "min")):
                    key = f"{min_max}_{var}"
                    func(val,
                         contingency_results[element].setdefault(key, np.full_like(val, np.nan, dtype=np.float64)),
                         out=contingency_results[element][key],
                         where=net[element]["in_service"].values & ~np.isnan(val))
            else:
                contingency_results[element][var] = val


def get_element_limits(net):
    """
    Construct the dictionary of element limits

    INPUT
    ----------
    **net** - pandapowerNet

    OUTPUT
    -------
    **element_limits** - dict
    """
    element_limits = {}
    if "max_vm_pu" in net.bus and "min_vm_pu" in net.bus:
        fill_val = np.full_like(net.bus.index, np.nan, np.float64)
        bus_index = net.bus.loc[~pd.isnull(net.bus.get("max_vm_pu", fill_val)) &
                                ~pd.isnull(net.bus.get("min_vm_pu", fill_val))].index.values
        if len(bus_index) != 0:
            element_limits.update({"bus": {
                "index": bus_index,
                "max_limit": net.bus.loc[bus_index, "max_vm_pu"].values,
                "min_limit": net.bus.loc[bus_index, "min_vm_pu"].values,
                "max_limit_nminus1":
                    net.bus.loc[bus_index, "max_vm_nminus1_pu"].values
                    if "max_vm_nminus1_pu" in net.bus.columns
                    else net.bus.loc[bus_index, "max_vm_pu"].values,
                "min_limit_nminus1":
                    net.bus.loc[bus_index, "min_vm_nminus1_pu"].values
                    if "min_vm_nminus1_pu" in net.bus.columns
                    else net.bus.loc[bus_index, "min_vm_pu"].values}})

    for element in ("line", "trafo", "trafo3w"):
        if net[element].empty or ("max_loading_percent" not in net[element] and
                                  "max_temperature_degree_celsius" not in net[element]):
            continue

        fill_val = np.full_like(net[element].index, np.nan, np.float64)
        element_index = net[element].loc[
            ~pd.isnull(net[element].get("max_loading_percent", fill_val)) |
            ~pd.isnull(net[element].get("max_temperature_degree_celsius", fill_val))].index.values
        if len(element_index) == 0:
            continue

        d = {element: {"index": element_index}}

        if "max_loading_percent" in net[element].columns:
            d[element].update({
                "max_limit": net[element].loc[element_index, "max_loading_percent"].values,
                "min_limit": -net[element].loc[element_index, "max_loading_percent"].values,
                "max_limit_nminus1":
                    net[element].loc[element_index, "max_loading_nminus1_percent"].values
                    if "max_loading_nminus1_percent" in net[element].columns
                    else net[element].loc[element_index, "max_loading_percent"].values,
                "min_limit_nminus1":
                    -net[element].loc[element_index, "max_loading_nminus1_percent"].values
                    if "max_loading_nminus1_percent" in net[element].columns
                    else - net[element].loc[element_index, "max_loading_percent"].values})

        if element == "line" and "max_temperature_degree_celsius" in net[element].columns:
            col = "max_temperature_degree_celsius"
            d[element].update({"max_temperature_degree_celsius": net.line.loc[element_index, col].values,
                               "min_temperature_degree_celsius": -net.line.loc[element_index, col].values})

        element_limits.update(d)
    return element_limits


def check_elements_within_limits(element_limits, contingency_results, nminus1=False, branch_tol=1e-3, bus_tol=1e-6):
    """
    Check if elements are within limits

    INPUT
    ----------
    **element_limits** - dict
    **contingency_results** - dict
    **nminus1** - bool
    **branch_tol** - float
        tolerance of the limit violation check for branch limits
    **bus_tol** - float
        tolerance of the limit violation check for bus limits

    OUTPUT
    -------
    True if all within limits (no violations), False if any limits violated
    """
    for element, values in contingency_results.items():
        limit = element_limits[element]
        if 'max_temperature_degree_celsius' in limit and 'temperature_degree_celsius' in values:
            if np.any(values['temperature_degree_celsius'] > limit['max_temperature_degree_celsius'] + branch_tol):
                return False
            if nminus1 and np.any(values['max_temperature_degree_celsius'] >
                                  limit['max_temperature_degree_celsius'] + branch_tol):
                return False
        if "max_limit" not in limit:
            continue
        if element == "bus":
            if np.any(values["vm_pu"] > limit["max_limit"] + bus_tol) or \
                    np.any(values["vm_pu"] < limit["min_limit"] - bus_tol):
                return False
            if nminus1 and (np.any(values["max_vm_pu"] > limit["max_limit_nminus1"] + bus_tol) or
                            np.any(values["min_vm_pu"] < limit["min_limit_nminus1"] - bus_tol)):
                return False
            continue
        if np.any(values['loading_percent'] > limit["max_limit"] + branch_tol):
            return False
        if nminus1 and np.any(values['max_loading_percent'] > limit["max_limit_nminus1"] + branch_tol):
            return False
    return True


def _get_sub_index(index, sub_index):
    lookup_index = np.zeros(index.max() + 1, dtype=np.int64)
    lookup_index[index] = np.arange(len(index), dtype=np.int64)
    return lookup_index[sub_index]


def _get_iloc_index(index, sub_index):
    # index_dict = {k: i for i, k in enumerate(index)}
    # return np.array([index_dict[i] for i in sub_index], dtype=np.int64)
    # should be the same:
    continuous_index = np.arange(len(index), dtype=np.int64)
    return continuous_index[_get_sub_index(index, sub_index)]


def _log_violation(element, var, val, limit_index, mask):
    if np.any(mask):
        s = ' (N-1)' if 'max' in var else ''
        with np.printoptions(precision=3, suppress=True):
            logger.info(f"{element}: {var}{s} violation at index {limit_index[mask]} ({val[mask]})")


def report_contingency_results(element_limits, contingency_results, branch_tol=1e-3, bus_tol=1e-6):
    """
    Print log messages for elements with violations of limits

    INPUT
    ----------
    **element_limits** - dict
    **contingency_results** - dict
    **branch_tol** - float
        tolerance for branch results
    **bus_tol** - float
        tolerance for bus results
    """
    for element, results in contingency_results.items():
        limit = element_limits[element]
        index = _get_sub_index(results["index"], limit["index"])
        for var, val in results.items():
            tol = bus_tol if element == "bus" else branch_tol
            if var == "index":
                continue
            if "min" in var:
                mask = val[index] < limit['min_limit_nminus1'] - tol
                _log_violation(element, var, val[index], limit["index"], mask)
            elif "max" in var:
                mask = val[index] > limit['max_limit_nminus1'] + tol
                _log_violation(element, var, val[index], limit["index"], mask)
            elif "cause" in var:
                continue
            else:
                mask_max = val[index] > limit['max_limit'] + tol
                _log_violation(element, var, val[index], limit["index"], mask_max)
                mask_min = val[index] < limit['min_limit'] - tol
                _log_violation(element, var, val[index], limit["index"], mask_min)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from pandapower.auxiliary import _detect_read_write_flag, write_to_net
from pandapower.control.basic_controller import Controller

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


class ConstControl(Controller):
    """
    Class representing a generic time series controller for a specified element and variable.
    Control strategy: "No Control" -> updates values of specified elements according to timeseries
    input data. If ConstControl is used without timeseries input data, it will reset the controlled
    values to the initial values, preserving the initial net state.
    The timeseries values are written to net during time_step before the initial powerflow run and
    before other controllers' control_step. It is possible to set attributes of objects that are
    contained in a net table, e.g. attributes of other controllers. This can be helpful
    e.g. if a voltage setpoint of a transformer tap changer depends on the time step.
    An attribute of an object in the "object" column of a
    table (e.g. net.controller["object"] -> net.controller.object.at[0, "vm_set_pu"]
    can be set if the attribute is specified as "object.attribute" (e.g. "object.vm_set_pu").

    INPUT:

        **net** (attrdict) - The net in which the controller resides

        **element** - element table ('sgen', 'load' etc.)

        **variable** - variable ('p_mw', 'q_mvar', 'vm_pu', 'tap_pos' etc.)

        **element_index** (int[]) - IDs of the controlled elements

        **data_source** (obj) - The data source that provides profile data

        **profile_name** (str[]) - The profile names of the elements in the data source


    OPTIONAL:

        **scale_factor** (real, 1.0) - Scaling factor for time series input values

        **in_service** (bool, True) - Indicates if the controller is currently in_service

        **recycle** (bool, True) - Re-use of internal-data in a time series loop.

        **drop_same_existing_ctrl** (bool, False) - Indicates if already existing controllers of
        the same type and with the same matching parameters (e.g. at same element) should be
        dropped

    .. note:: If multiple elements are represented with one controller, the data source must have
        integer columns. At the moment, only the DFData format is tested for the multiple const
        control.
    """

    def __init__(self, net, element, variable, element_index, profile_name=None, data_source=None,
                 scale_factor=1.0, in_service=True, recycle=True, order=-1, level=-1,
                 drop_same_existing_ctrl=False, matching_params=None,
                 initial_run=False, **kwargs):
        # just calling init of the parent
        if matching_params is None:
            matching_params = {"element": element, "variable": variable,
                               "element_index": element_index}
        super().__init__(net, in_service=in_service, recycle=recycle, order=order, level=level,
                         drop_same_existing_ctrl=drop_same_existing_ctrl,
                         matching_params=matching_params, initial_run=initial_run,
                         **kwargs)

        # data source for time series values
        self.data_source = data_source
        # ids of sgens or loads
        self.element_index = element_index
        # element type
        self.element = element
        self.values = None
        self.profile_name = profile_name
        self.scale_factor = scale_factor
        self.applied = False
        self.write_flag, self.variable = _detect_read_write_flag(
            net, element, element_index, variable)
        self.set_recycle(net)

    def set_recycle(self, net):
        allowed_elements = ["load", "sgen", "storage", "gen", "ext_grid", "trafo", "trafo3w", "line"]
        if net.controller.at[self.index, 'recycle'] is False or self.element not in allowed_elements:
            # if recycle is set to False by the user when creating the controller it is deactivated
            # or when const control controls an element which is not able to be recycled
            net.controller.at[self.index, 'recycle'] = False
            return
        # these variables determine what is re-calculated during a time series run
        recycle = dict(trafo=False, gen=False, bus_pq=False)
        if self.element in ["sgen", "load", "storage"] and self.variable in ["p_mw", "q_mvar",
                                                                             "scaling"]:
            recycle["bus_pq"] = True
        if self.element in ["gen"] and self.variable in ["p_mw", "vm_pu", "scaling"] \
                or self.element in ["ext_grid"] and self.variable in ["vm_pu", "va_degree"]:
            recycle["gen"] = True
        if self.element in ["trafo", "trafo3w", "line"]:
            recycle["trafo"] = True
        # recycle is either the dict what should be recycled
        # or False if the element + variable combination is not supported
        net.controller.at[self.index, 'recycle'] = recycle if any(list(recycle.values())) else False

    def time_step(self, net, time):
        """
        Get the values of the element from data source
        Write to pandapower net by calling write_to_net()
        If ConstControl is used without a data_source, it will reset the controlled values to the
        initial values,
        preserving the initial net state.
        """
        self.applied = False
        if self.data_source is None:
            self.values = net[self.element][self.variable].loc[self.element_index]
        else:
            self.values = self.data_source.get_time_step_value(time_step=time,
                                                               profile_name=self.profile_name,
                                                               scale_factor=self.scale_factor)
        if self.values is not None:
            write_to_net(net, self.element, self.element_index, self.variable, self.values,
                         self.write_flag)

    def is_converged(self, net):
        """
        Actual implementation of the convergence criteria: If controller is applied, it can stop
        """
        return self.applied

    def control_step(self, net):
        """
        Set applied to True, which means that the values set in time_step have been included in the
        load flow calculation.
        """
        self.applied = True

    def __str__(self):
        return super().__str__() + " [%s.%s]" % (self.element, self.variable)



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import logging
import datetime
import os
import tempfile
import time
from typing import Dict, List
import numpy as np
import pandas as pd


class UCTEParser:
    def __init__(self, path_ucte_file: str = None, config: Dict = None):
        """
        This class parses an UCTE file and loads its content to a dictionary of
        UCTE element type (str) -> UCTE elements (DataFrame)
        :param path_ucte_file: The path to the UCTE file. Optional, default: None. This parameter may be set later.
        :param config: The configuration dictionary. Optional, default: None. This parameter may be set later.
        """
        self.path_ucte_file: str = path_ucte_file
        self.config: Dict = config if isinstance(config, dict) else dict()
        self.logger = logging.getLogger(self.__class__.__name__)
        self.ucte_elements = ["##C", "##N", "##L", "##T", "##R", "##TT", "##E"]
        self.data: Dict[str, pd.DataFrame] = dict()
        self.date: datetime.datetime = datetime.datetime.now(datetime.timezone.utc)
        self.bus_ucte_countries : list = list()

    def parse_file(self, path_ucte_file: str = None) -> bool:
        """
        Parse a UCTE file.
        :param path_ucte_file: The path to the UCTE file. By default the internal field 'self.path_ucte_file' is used.
        :return: True if parsing was successful, False otherwise.
        """
        time_start = time.time()
        # check parameters and needed fields
        if path_ucte_file is not None:
            self.path_ucte_file = path_ucte_file
        if self.path_ucte_file is None:
            self.logger.error(
                "The variable 'path_ucte_file' is None, set it before parsing it!"
            )
            return False
        self.logger.info("Start parsing the file %s." % self.path_ucte_file)
        # parse the date, if its given by the configuration, parse it from there, otherwise from the file string
        if (
            "custom" in self.config.keys()
            and isinstance(self.config["custom"], dict)
            and "date" in self.config["custom"].keys()
        ):
            self._parse_date_str(self.config["custom"]["date"])
        else:
            self._parse_date_str(os.path.basename(self.path_ucte_file)[:13])
        raw_input_dict = dict()
        for ucte_element in self.ucte_elements:
            raw_input_dict[ucte_element] = []
        with open(self.path_ucte_file, "r") as f:
            # current_element contains the UCTE element type which is actually in parse progress, e.g. '##N'
            current_element = ""
            ucte_country = ""
            is_in_N = False
            # iterate through the origin input file
            for row in f.readlines():
                row = row.strip()
                if row.startswith("##N"):
                    is_in_N = True
                elif any([row.startswith(other_ucte_element) for other_ucte_element in
                          self.ucte_elements if other_ucte_element != "##N"]):
                    is_in_N = False
                if row in self.ucte_elements:
                    # the start of a new UCTE element type in the origin file
                    current_element = row
                elif row.startswith("##C"):
                    # special for comments, because '##C' rows look like '##C 2007.05.01'
                    current_element = "##C"
                    raw_input_dict[current_element].append(row)
                elif row.startswith("##"):
                    self.logger.debug("Skipping row %s" % row)
                    ucte_country = row[3:] if row[2] == "Z" else row[2:] # usually country zones are marked by ##Z
                else:
                    if is_in_N:
                        self.bus_ucte_countries.append(ucte_country)
                    raw_input_dict[current_element].append(row)
        self._create_df_from_raw(raw_input_dict)
        self.logger.info(
            "Finished parsing file %s in %ss."
            % (self.path_ucte_file, time.time() - time_start)
        )
        return True

    def _parse_date_str(self, date_str: str):
        try:
            self.date = datetime.datetime.strptime(date_str, "%Y%m%d_%H%M")
        except Exception as e:
            self.logger.info(
                f"The given {date_str=} couldn't be parsed as '%Y%m%d_%H%M'.")
            self.date = datetime.datetime.utcnow()

    def _create_df_from_raw(self, raw_input_dict):
        # create DataFrames from the raw_input_dict
        self.data = dict()
        for ucte_element, items in raw_input_dict.items():
            self.data[ucte_element] = pd.DataFrame(items)
        # make sure that at least some empty data exist
        if 0 not in self.data["##C"].columns:
            self.data["##C"][0] = ""
        # split the raw input for each UCTE element type
        self.data["##C"]["comments"] = self.data["##C"][0][:]
        self._split_nodes_from_raw()
        self._split_lines_from_raw()
        self._split_trafos_from_raw()
        self._split_trafos_regulation_from_raw()
        self._split_trafos_specified_parameters_from_raw()
        self._split_exchange_powers_from_raw()
        # drop the raw input columns
        for ucte_element, df in self.data.items():
            if 0 in df.columns:
                df = df.drop(columns=[0], axis=1)
        # set the data types
        dtypes = dict()
        i_t = pd.Int64Dtype()
        dtypes["##N"] = dict(
            {
                "status": i_t,
                "voltage": float,
                "p_load": float,
                "q_load": float,
                "p_gen": float,
                "q_gen": float,
                "min_p_gen": float,
                "max_p_gen": float,
                "min_q_gen": float,
                "max_q_gen": float,
                "static_primary_control": float,
                "p_primary_control": float,
                "three_ph_short_circuit_power": float,
                "x_r_ratio": float,
                "node_type": i_t,
            }
        )
        dtypes["##L"] = dict(
            {"status": i_t, "r": float, "x": float, "b": float, "i": float}
        )
        dtypes["##T"] = dict(
            {
                "status": i_t,
                "voltage1": float,
                "voltage2": float,
                "s": float,
                "r": float,
                "x": float,
                "b": float,
                "g": float,
                "i": float,
            }
        )
        dtypes["##R"] = dict(
            {
                "phase_reg_delta_u": float,
                "phase_reg_n": float,
                "phase_reg_n2": float,
                "phase_reg_u": float,
                "angle_reg_delta_u": float,
                "angle_reg_theta": float,
                "angle_reg_n": float,
                "angle_reg_n2": float,
                "angle_reg_p": float,
            }
        )
        dtypes["##TT"] = dict(
            {
                "tap_position": float,
                "r": float,
                "x": float,
                "delta_u": float,
                "alpha": float,
            }
        )
        dtypes["##E"] = dict({"p": float})
        for ucte_element, one_dtypes in dtypes.items():
            for field, field_type in one_dtypes.items():
                self.data[ucte_element].loc[
                    self.data[ucte_element][field] == "", field
                ] = np.nan
                # for integer: first convert them to float to prevent errors
                if field_type == i_t:
                    self.data[ucte_element][field] = self.data[ucte_element][
                        field
                    ].astype(float)
                self.data[ucte_element][field] = self.data[ucte_element][field].astype(
                    field_type
                )

        # remove '##' at the beginning of each key
        for one_key in list(self.data.keys()):
            self.data[one_key[2:]] = self.data[one_key]
            self.data.pop(one_key)

        del raw_input_dict

    def _split_nodes_from_raw(self):
        element_type = "##N"
        if element_type not in self.data.keys():
            self.logger.warning("No nodes in 'self.data' available! Didn't split them.")
            return
        df = self.data[element_type]
        # if 0 not in df.columns:
        #     df[0] = ""
        df["node"] = df[0].str[0:8].str.strip()
        df["node_name"] = df[0].str[9:21].str.strip()
        df["status"] = df[0].str[22:23].str.strip()
        df["node_type"] = df[0].str[24:25].str.strip()
        df["voltage"] = df[0].str[26:32].str.strip()
        df["p_load"] = df[0].str[33:40].str.strip()
        df["q_load"] = df[0].str[41:48].str.strip()
        df["p_gen"] = df[0].str[49:56].str.strip()
        df["q_gen"] = df[0].str[57:64].str.strip()
        df["min_p_gen"] = df[0].str[65:72].str.strip()
        df["max_p_gen"] = df[0].str[73:80].str.strip()
        df["min_q_gen"] = df[0].str[81:88].str.strip()
        df["max_q_gen"] = df[0].str[89:96].str.strip()
        df["static_primary_control"] = df[0].str[97:102].str.strip()
        df["p_primary_control"] = df[0].str[103:110].str.strip()
        df["three_ph_short_circuit_power"] = df[0].str[111:118].str.strip()
        df["x_r_ratio"] = df[0].str[119:126].str.strip()
        df["type"] = df[0].str[127:128].str.strip()
        df["ucte_country"] = self.bus_ucte_countries

    def _split_connections_from_raw(self, element_type: str):
        # get the connections to nodes (node1 and node2) from the asset for e.g. ##L or ##T
        self.data[element_type]["node1"] = (
            self.data[element_type][0].str[0:8].str.strip()
        )
        self.data[element_type]["node2"] = (
            self.data[element_type][0].str[9:17].str.strip()
        )

    def _split_lines_from_raw(self):
        element_type = "##L"
        if element_type not in self.data.keys():
            self.logger.warning("No lines in 'self.data' available! Didn't split them.")
            return
        df = self.data[element_type]
        if 0 not in df.columns:
            df[0] = ""
        self._split_connections_from_raw(element_type)
        df["order_code"] = df[0].str[18:19].str.strip()
        df["status"] = df[0].str[20:21].str.strip()
        df["r"] = df[0].str[22:28].str.strip()
        df["x"] = df[0].str[29:35].str.strip()
        df["b"] = df[0].str[36:44].str.strip()
        df["i"] = df[0].str[45:51].str.strip()
        df["name"] = df[0].str[52:64].str.strip()

    def _split_trafos_from_raw(self):
        element_type = "##T"
        if element_type not in self.data.keys():
            self.logger.warning(
                "No transformers in 'self.data' available! Didn't split them."
            )
            return
        df = self.data[element_type]
        if 0 not in df.columns:
            df[0] = ""
        self._split_connections_from_raw(element_type)
        df["order_code"] = df[0].str[18:19].str.strip()
        df["status"] = df[0].str[20:21].str.strip()
        df["voltage1"] = df[0].str[22:27].str.strip()
        df["voltage2"] = df[0].str[28:33].str.strip()
        df["s"] = df[0].str[34:39].str.strip()
        df["r"] = df[0].str[40:46].str.strip()
        df["x"] = df[0].str[47:53].str.strip()
        df["b"] = df[0].str[54:62].str.strip()
        df["g"] = df[0].str[63:69].str.strip()
        df["i"] = df[0].str[70:76].str.strip()
        df["name"] = df[0].str[77:89].str.strip()

    def _split_trafos_regulation_from_raw(self):
        element_type = "##R"
        if element_type not in self.data.keys():
            self.logger.warning(
                "No tap changers in 'self.data' available! Didn't split them."
            )
            return
        df = self.data[element_type]
        if 0 not in df.columns:
            df[0] = ""
        self._split_connections_from_raw(element_type)
        df["order_code"] = df[0].str[18:19].str.strip()
        df["phase_reg_delta_u"] = df[0].str[20:25].str.strip()
        df["phase_reg_n"] = df[0].str[26:28].str.strip()
        df["phase_reg_n2"] = df[0].str[29:32].str.strip()
        df["phase_reg_u"] = df[0].str[33:38].str.strip()
        df["angle_reg_delta_u"] = df[0].str[39:44].str.strip()
        df["angle_reg_theta"] = df[0].str[45:50].str.strip()
        df["angle_reg_n"] = df[0].str[51:53].str.strip()
        df["angle_reg_n2"] = df[0].str[54:57].str.strip()
        df["angle_reg_p"] = df[0].str[58:63].str.strip()
        df["angle_reg_type"] = df[0].str[64:68].str.strip()

    def _split_trafos_specified_parameters_from_raw(self):
        element_type = "##TT"
        if element_type not in self.data.keys():
            self.logger.warning(
                "No specified transformer parameters in 'self.data' available! Didn't split them."
            )
            return
        df = self.data[element_type]
        if 0 not in df.columns:
            df[0] = ""
        self._split_connections_from_raw(element_type)
        df["order_code"] = df[0].str[18:19].str.strip()
        df["tap_position"] = df[0].str[22:25].str.strip()
        df["r"] = df[0].str[26:32].str.strip()
        df["x"] = df[0].str[33:39].str.strip()
        df["delta_u"] = df[0].str[40:45].str.strip()
        df["alpha"] = df[0].str[46:51].str.strip()

    def _split_exchange_powers_from_raw(self):
        element_type = "##E"
        if element_type not in self.data.keys():
            self.logger.warning(
                "No exchange powers in 'self.data' available! Didn't split them."
            )
            return
        df = self.data[element_type]
        if 0 not in df.columns:
            df[0] = ""
        df["country1"] = df[0].str[0:2].str.strip()
        df["country2"] = df[0].str[3:5].str.strip()
        df["p"] = df[0].str[6:13].str.strip()
        df["comment"] = df[0].str[14:26].str.strip()

    def get_fields_dict(self) -> Dict[str, Dict[str, List[str]]]:
        """
        Get an overview about the UCTE element types and their fields.
        :return: A dictionary containing a dictionary with the UCTE element types as keys and for each UCTE element
        type a list of fields and a dictionary with the UCTE element types as keys and for each UCTE element type a
        list of dtypes.
        """
        self.logger.info(
            "Creating a dictionary containing the UCTE element types as keys and for each UCTE element "
            "type a list of fields and dtypes."
        )
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.close()
            ucte_temp = UCTEParser()
            ucte_temp.set_config(dict({"custom": {"date": "20200701_1010"}}))
            ucte_temp.parse_file(path_ucte_file=f.name)
            data = ucte_temp.get_data()
        if os.path.exists(f.name):
            os.remove(f.name)
        return_dict = dict()
        return_dict["element_types"] = dict()
        return_dict["dtypes"] = dict()
        for element_type, df in data.items():
            return_dict["element_types"][element_type] = list(df.columns)
            return_dict["dtypes"][element_type] = [str(x) for x in df.dtypes.values]
        return return_dict

    def set_path_ucte_file(self, path_ucte_file: str):
        self.path_ucte_file = path_ucte_file

    def get_path_ucte_file(self) -> str:
        return self.path_ucte_file

    def set_config(self, config: Dict):
        if isinstance(config, dict):
            self.config = config
        else:
            self.logger.warning(
                "The configuration is not a dictionary! Default configuration is set."
            )
            self.config = dict()

    def get_config(self) -> Dict:
        return self.config

    def get_data(self) -> Dict[str, pd.DataFrame]:
        return self.data

    def get_date(self) -> datetime.datetime:
        return self.date



def prj_import(com_import, dst_user_dir, file_name):
    com_import.g_target = dst_user_dir
    com_import.g_file = file_name
    com_import.g_target = dst_user_dir
    assert com_import.g_target is dst_user_dir
    com_import.Execute()


def prj_dgs_import(com_import, dst_user_dir, file_name, name, template=None):
    com_import.targpath = dst_user_dir
    com_import.targname = name
    com_import.fFile = file_name
    if template is not None:
        com_import.prjTemplate = template
    com_import.Execute()


def choose_imp_dir(user, IMPFOLD):
    imp_dir_list = user.GetContents(IMPFOLD)
    if len(imp_dir_list) != 0:
        imp_dir = imp_dir_list[0]
    else:
        imp_dir = user.CreateObject('IntFolder', IMPFOLD)
    return imp_dir


def clear_dir(dir):
    trash = dir.GetContents()
    for item in trash:
        item.Delete()


def run_ldf(com_ldf):
    com_ldf.SetAttribute('iopt_net', 0)
    # com_ldf.SetAttribute('iopt_at', 1)
    com_ldf.SetAttribute('iopt_pq', 0)

    com_ldf.SetAttribute('errlf', 0.001)
    com_ldf.SetAttribute('erreq', 0.01)

    com_ldf.Execute()



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np
from scipy.optimize import linprog

# Make sure the library is installed
# Otherwise, load the scipy linprog
try:
    from ortools.linear_solver import pywraplp
    ortools_available = True
except ImportError:
    ortools_available = False
import warnings

from pandapower.estimation.ppc_conversion import ExtendedPPCI
from pandapower.estimation.algorithm.base import BaseAlgorithm
from pandapower.estimation.algorithm.matrix_base import BaseAlgebra


class LPAlgorithm(BaseAlgorithm):
    def estimate(self, eppci: ExtendedPPCI, with_ortools=True, **kwargs):
        if "estimator" in kwargs and kwargs["estimator"].lower() != "lav":  # pragma: no cover
            self.logger.warning("LP Algorithm supports only LAV Estimator!! Set to LAV!!")

        # matrix calculation object
        sem = BaseAlgebra(eppci)

        current_error, cur_it = 100., 0
        E = eppci.E
        while current_error > self.tolerance and cur_it < self.max_iterations:
            self.logger.debug("Starting iteration {:d}".format(1 + cur_it))
            try:
                # residual r
                r = sem.create_rx(E)

                # jacobian matrix H
                H = sem.create_hx_jacobian(E)

                # state vector difference d_E
                # d_E = G_m^-1 * (H' * R^-1 * r)
                d_E = self._solve_lp(H, E, r, with_ortools=with_ortools)
                E += d_E
                eppci.update_E(E)

                # prepare next iteration
                cur_it += 1
                current_error = np.max(np.abs(d_E))
                self.logger.debug("Current error: {:.7f}".format(current_error))
            except np.linalg.linalg.LinAlgError:  # pragma: no cover
                self.logger.error("A problem appeared while using the linear algebra methods."
                                  "Check and change the measurement set.")
                return False

        # check if the estimation is successfull
        self.check_result(current_error, cur_it)
        return eppci

    @staticmethod
    def _solve_lp(H, x, r, with_ortools):

        """Function to choose the best option based on the installed libraries to solve linear programming.

        Performance comparison(601 bus system/1204 random measurements):
        Scipy   : 269.20 seconds
        OR-Tools:   8.51 seconds +- 154 ms
        """

        n, m = H.shape[1], H.shape[0]
        zero_n = np.zeros((n, 1))
        one_m = np.ones((m, 1))
        Im = np.eye(m)

        c_T = np.r_[zero_n, zero_n, one_m, one_m]
        A = np.c_[H, -H, Im, -Im]

        if ortools_available and with_ortools:
            return LPAlgorithm._solve_or_tools(c_T.ravel(), A, r, n)
        else:
            return LPAlgorithm._solve_scipy(c_T.ravel(), A, r, n)

    @staticmethod
    def _solve_scipy(c_T, A, r, n):

        """The use of linprog function from the scipy library."""

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            res = linprog(c_T.ravel(), A_eq=A, b_eq=r,
                          method="simplex", options={'tol': 1e-5, 'disp': False, 'maxiter': 20000})
        if res.success:
            d_x = np.array(res['x'][:n]).ravel() - np.array(res['x'][n:2 * n]).ravel()
            return d_x
        else:  # pragma: no cover
            raise np.linalg.linalg.LinAlgError


    @staticmethod
    def _solve_or_tools(c_T, A, r, n):
        # Just to ensure floating point precision if there are any
        error_margin = 1e-10

        #'GLOP' fails with ortools version > 9.4.1874
        solver = pywraplp.Solver.CreateSolver('SCIP')


        # Create the states...
        x = []
        for counter in range(len(c_T)):
            x.append(solver.NumVar(0, solver.infinity(), 'x_' + str(counter)))


        # Give the equality constraints...
        for row_counter, row in enumerate(A):
            row_equality = 0
            for col_counter, col in enumerate(row):
                if abs(col) > error_margin:
                    row_equality = row_equality + col*x[col_counter]

            solver.Add(row_equality == r[row_counter])


        # What to optimize?
        to_minimize = 0
        for counter, coef in enumerate(c_T):
            to_minimize = to_minimize + coef*x[counter]

        solver.Minimize(to_minimize)

        # Solve the optimization problem
        status = solver.Solve()


        d_x = []
        if status == pywraplp.Solver.OPTIMAL or status == solver.FEASIBLE:
            # An optimal solution is found
            for counter, x_current in enumerate(x):
                d_x.append(x_current.solution_value())

            d_x = np.array(d_x)
            return d_x[:n] - d_x[n:2*n]
        else:
            # No solution found...
            raise np.linalg.linalg.LinAlgError





# -*- coding: utf-8 -*-
# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.
import logging
import os
import json
from typing import Dict, List
import numpy as np
from pandapower.auxiliary import pandapowerNet
import pandas as pd

logger = logging.getLogger(__name__)


def get_pp_net_special_columns_dict() -> Dict[str, str]:
    """
    Get a dictionary with the special CIM fields, used as columns in a pandapower network.
    :return Dict[str, str]: The dictionary with the special CIM fields.
    """
    return dict({'o_id': 'origin_id', 'sub': 'substation', 't': 'terminal', 't_from': 'terminal_from',
                 'from_bus': 'terminal_from', 't_to': 'terminal_to', 'to_bus': 'terminal_to', 't_bus': 'terminal_bus',
                 't_ele': 'terminal_element', 't_hv': 'terminal_hv', 'hv_bus': 'terminal_hv', 't_mv': 'terminal_mv',
                 'mv_bus': 'terminal_mv', 't_lv': 'terminal_lv', 'lv_bus': 'terminal_lv', 'o_cl': 'origin_class',
                 'o_prf': 'origin_profile', 'ct': 'cim_topnode', 'tc': 'tapchanger_class', 'tc_id': 'tapchanger_id',
                 'pte_id': 'PowerTransformerEnd_id', 'pte_id_hv': 'PowerTransformerEnd_id_hv',
                 'pte_id_mv': 'PowerTransformerEnd_id_mv', 'pte_id_lv': 'PowerTransformerEnd_id_lv',
                 'cnc_id': 'ConnectivityNodeContainer_id', 'sub_id': 'Substation_id', 'src': 'source', 'name': 'name',
                 'desc': 'description', 'a_id': 'analog_id', 'bus': 'terminal'})


def extend_pp_net_cim(net: pandapowerNet, override: bool = True) -> pandapowerNet:
    """
    Extend pandapower element DataFrames with special columns for the CIM converter, e.g. a column for the RDF ID.
    :param net: The pandapower net to extend.
    :param override: If True, all existing special CIM columns will be overwritten (content will be erased). If False,
    only missing columns will be created. Optional, default: True
    :return: The extended pandapower network.
    """
    np_str_type = 'str'
    np_float_type = 'float'
    np_bool_type = 'bool'

    sc = get_pp_net_special_columns_dict()

    # all pandapower element types like bus, line, trafo will get the following special columns
    fill_dict_all: Dict[str, List[str]] = dict({})
    fill_dict_all[np_str_type] = [sc['o_id'], sc['o_cl']]

    # special elements
    fill_dict: Dict[str, Dict[str, List[str]]] = dict()

    fill_dict['bus'] = dict()
    fill_dict['bus'][np_str_type] = [sc['o_prf'], sc['ct'], sc['cnc_id'], sc['sub_id'], 'description', 'Busbar_id',
                                     'Busbar_name', 'GeographicalRegion_id', 'GeographicalRegion_name',
                                     'SubGeographicalRegion_id', 'SubGeographicalRegion_name']

    fill_dict['ext_grid'] = dict()
    fill_dict['ext_grid'][np_str_type] = [sc['t'], sc['sub'], 'description']
    fill_dict['ext_grid'][np_float_type] = ['min_p_mw', 'max_p_mw', 'min_q_mvar', 'max_q_mvar', 'p_mw', 'q_mvar',
                                            's_sc_max_mva', 's_sc_min_mva', 'rx_max', 'rx_min', 'r0x0_max', 'x0x_max']

    fill_dict['load'] = dict()
    fill_dict['load'][np_str_type] = [sc['t'], 'description']
    fill_dict['gen'] = dict()
    fill_dict['gen'][np_str_type] = [sc['t'], 'description']
    fill_dict['gen'][np_float_type] = \
        ['min_p_mw', 'max_p_mw', 'min_q_mvar', 'max_q_mvar', 'vn_kv', 'rdss_ohm', 'xdss_pu', 'cos_phi', 'pg_percent', 'governorSCD']
    fill_dict['sgen'] = dict()
    fill_dict['sgen'][np_str_type] = [sc['t'], 'description', 'generator_type']
    fill_dict['sgen'][np_float_type] = ['k', 'rx', 'vn_kv', 'rdss_ohm', 'xdss_pu', 'lrc_pu']
    fill_dict['motor'] = dict()
    fill_dict['motor'][np_str_type] = [sc['t'], 'description']
    fill_dict['storage'] = dict()
    fill_dict['storage'][np_str_type] = [sc['t'], 'description']
    fill_dict['shunt'] = dict()
    fill_dict['shunt'][np_str_type] = [sc['t'], 'description','sVCControlMode']
    fill_dict['ward'] = dict()
    fill_dict['ward'][np_str_type] = [sc['t'], 'description']
    fill_dict['xward'] = dict()
    fill_dict['xward'][np_str_type] = [sc['t'], 'description']

    fill_dict['line'] = dict()
    fill_dict['line'][np_str_type] = [sc['t_from'], sc['t_to'], 'description', 'EquipmentContainer_id']
    fill_dict['line'][np_float_type] = ['r0_ohm_per_km', 'x0_ohm_per_km', 'c0_nf_per_km', 'g0_us_per_km',
                                        'endtemp_degree']

    fill_dict['dcline'] = dict()
    fill_dict['dcline'][np_str_type] = [sc['t_from'], sc['t_to'], 'description']

    fill_dict['switch'] = dict()
    fill_dict['switch'][np_str_type] = [sc['t_bus'], sc['t_ele'], 'description']

    fill_dict['impedance'] = dict()
    fill_dict['impedance'][np_str_type] = [sc['t_from'], sc['t_to'], 'description']
    fill_dict['impedance'][np_float_type] = ['rft0_pu', 'xft0_pu', 'rtf0_pu', 'xtf0_pu']

    fill_dict['trafo'] = dict()
    fill_dict['trafo'][np_str_type] = [sc['t_hv'], sc['t_lv'], sc['pte_id_hv'], sc['pte_id_lv'], sc['tc'], sc['tc_id'],
                                       'description', 'vector_group']
    fill_dict['trafo'][np_float_type] = ['vk0_percent', 'vkr0_percent', 'xn_ohm']
    fill_dict['trafo'][np_bool_type] = ['power_station_unit', 'oltc']

    fill_dict['trafo3w'] = dict()
    fill_dict['trafo3w'][np_str_type] = [sc['t_hv'], sc['t_mv'], sc['t_lv'], sc['pte_id_hv'], sc['pte_id_mv'],
                                         sc['pte_id_lv'], sc['tc'], sc['tc_id'], 'description', 'vector_group']
    fill_dict['trafo3w'][np_float_type] = ['vk0_hv_percent', 'vk0_mv_percent', 'vk0_lv_percent', 'vkr0_hv_percent',
                                           'vkr0_mv_percent', 'vkr0_lv_percent']
    fill_dict['trafo3w'][np_bool_type] = ['power_station_unit']

    fill_dict['measurement'] = dict()
    fill_dict['measurement'][np_str_type] = ['source', 'origin_class', 'origin_id', 'analog_id', 'terminal_id',
                                             'description']

    for pp_type, one_fd in fill_dict.items():
        for np_type, fields in fill_dict_all.items():
            np_type = np.sctypeDict.get(np_type)
            for field in fields:
                if override or field not in net[pp_type].columns:
                    net[pp_type][field] = pd.Series([], dtype=np_type)
        for np_type, fields in one_fd.items():
            np_type = np.sctypeDict.get(np_type)
            for field in fields:
                if override or field not in net[pp_type].columns:
                    net[pp_type][field] = pd.Series([], dtype=np_type)

    # some special items
    if override:
        net['CGMES'] = dict()
        net['CGMES']['BaseVoltage'] = pd.DataFrame(None, columns=['rdfId', 'nominalVoltage'])
    else:
        if 'CGMES' not in net.keys():
            net['CGMES'] = dict()
        if 'BaseVoltage' not in net['CGMES'].keys():
            net['CGMES']['BaseVoltage'] = pd.DataFrame(None, columns=['rdfId', 'nominalVoltage'])

    return net


def get_cim_schema(cgmes_version: str = '2.4.15') -> Dict[str, Dict[str, Dict[str, str or Dict[str, Dict[str, str]]]]]:
    """
    Parses the CIM schema from the serialized CIM schema json files which have been created from the RDF schema files.
    The schema is parsed for the serializer from the CIM data structure used by the cim2pp and pp2cim converters.

    :param cgmes_version: CIM version to use, '2.4.15' or '3.0', default '2.4.15'
    :return: The CIM schema as dictionary.
    """
    path_with_serialized_schemas = os.path.dirname(__file__) + os.sep + 'serialized_schemas'
    if not os.path.isdir(path_with_serialized_schemas):
        os.mkdir(path_with_serialized_schemas)
    for one_file in os.listdir(path_with_serialized_schemas):
        path_to_schema = path_with_serialized_schemas + os.sep + one_file
        if one_file.lower().startswith('cim16_') and cgmes_version == '2.4.15':
            logger.info("Parsing the schema from CIM 16 from disk: %s" % path_to_schema)
            with open(path_to_schema, encoding='UTF-8', mode='r') as f:
                cim_schema = json.load(f)
            return cim_schema
        elif one_file.lower().startswith('cim100_') and cgmes_version == '3.0':
            logger.info("Parsing the schema from CIM 100 from disk: %s" % path_to_schema)
            with open(path_to_schema, encoding='UTF-8', mode='r') as f:
                cim_schema = json.load(f)
            return cim_schema
